"""Backends for optical simulations.

This module provides a unified interface for different optical simulation backends,
as well as functions to interact with these backends.

Interfaces:

- `BaseAnalysisRegistry`: Base class for the backend analysis registry.
- `BaseBackend`: Base class for simulation backends.

Functions:

- `set_backend`: Set the backend to use for optical simulations.
- `get_backend`: Get the current backend, or initialize the default backend if not set.
- `get_oss`: Get the OpticStudioSystem instance if the current backend is OpticStudio.
- `get_optic`: Get the Optic instance if the current backend is Optiland.
- `update_settings`: Update settings on the current backend.

See Also
--------
visispy.opticstudio.backend : Backend for OpticStudio.
visispy.optiland.backend : Backend for Optiland.
"""

from __future__ import annotations

import json
from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from enum import Enum
from pathlib import Path
from types import MethodType
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Literal, TypeVar, cast, overload
from warnings import warn

from visisipy.types import NotRequired, TypedDict, Unpack

if TYPE_CHECKING:
    from collections.abc import Iterable
    from os import PathLike

    from optiland.optic import Optic
    from pandas import DataFrame
    from zospy.zpcore import OpticStudioSystem

    from visisipy.analysis.cardinal_points import CardinalPointsResult
    from visisipy.models import BaseEye, EyeModel
    from visisipy.refraction import FourierPowerVectorRefraction
    from visisipy.types import ApertureType, FieldCoordinate, FieldType, SampleSize
    from visisipy.wavefront import ZernikeCoefficients

__all__ = (
    "Backend",
    "BackendSettings",
    "BaseBackend",
    "get_backend",
    "get_optic",
    "get_oss",
    "set_backend",
    "update_settings",
)


_BACKEND: type[BaseBackend] | None = None
_DEFAULT_BACKEND: Backend | Literal["opticstudio", "optiland"] = "opticstudio"


_Analysis = TypeVar("_Analysis", bound=Callable)


class _AnalysisMethod(Generic[_Analysis]):
    def __init__(self, analysis) -> None:
        self._analysis = analysis

    @overload
    def __get__(self, instance: None, owner: type[BaseAnalysisRegistry]) -> _AnalysisMethod[_Analysis]: ...

    @overload
    def __get__(self, instance: BaseAnalysisRegistry, owner: type[BaseAnalysisRegistry]) -> _Analysis: ...

    def __get__(
        self, instance: BaseAnalysisRegistry | None, owner: type[BaseAnalysisRegistry]
    ) -> _AnalysisMethod | _Analysis:
        if instance is None:
            return self

        return MethodType(self._analysis, instance.backend)


class _classproperty(property):  # noqa: N801
    def __get__(self, instance, owner=None):
        return self.fget(owner)


class BaseAnalysisRegistry(ABC):
    """Base class for analysis registry.

    Interface for the analysis methods of the backend. Backends should implement this interface
    including all the analysis methods. If an analysis method is not implemented in the backend,
    it should raise a NotImplementedError.

    Attributes
    ----------
    backend : BaseBackend
        The backend in which the analysis is performed.

    Methods
    -------
    cardinal_points(surface_1, surface_2)
        Calculate the cardinal points of the optical system.
    fft_psf(field_coordinate, wavelength, field_type, sampling)
        Calculate the FFT Point Spread Function (PSF) at the retina surface.
    huygens_psf(field_coordinate, wavelength, field_type, pupil_sampling, image_sampling)
        Calculate the Huygens Point Spread Function (PSF) at the retina surface.
    raytrace(coordinates, wavelengths, field_type, pupil)
        Perform a raytrace through the optical system.
    refraction(field_coordinate, wavelength, sampling, pupil_diameter, field_type)
        Calculate the spherical equivalent of refraction for the optical system.
    strehl_ratio(field_coordinate, wavelength, field_type, sampling, psf_type)
        Calculate the Strehl ratio of the optical system.
    zernike_standard_coefficients(field_coordinate, wavelength, field_type, sampling, maximum_term)
        Calculate the Zernike standard coefficients for the optical system.
    """

    def __init__(self, backend: BaseBackend) -> None:
        self._backend = backend

    @property
    def backend(self) -> BaseBackend:
        return self._backend

    @abstractmethod
    def cardinal_points(
        self,
        surface_1: int | None = None,
        surface_2: int | None = None,
    ) -> tuple[CardinalPointsResult, Any]: ...

    @abstractmethod
    def fft_psf(
        self,
        field_coordinate: FieldCoordinate | None = None,
        wavelength: float | None = None,
        field_type: FieldType = "angle",
        sampling: SampleSize | str | int = 128,
    ) -> tuple[DataFrame, Any]: ...

    @abstractmethod
    def huygens_psf(
        self,
        field_coordinate: FieldCoordinate | None = None,
        wavelength: float | None = None,
        field_type: FieldType = "angle",
        pupil_sampling: SampleSize | str | int = 128,
        image_sampling: SampleSize | str | int = 128,
    ) -> tuple[DataFrame, Any]: ...

    @abstractmethod
    def raytrace(
        self,
        coordinates: Iterable[FieldCoordinate] | None = None,
        wavelengths: Iterable[float] | None = None,
        field_type: FieldType = "angle",
        pupil: tuple[float, float] = (0, 0),
    ) -> tuple[DataFrame, Any]: ...

    @abstractmethod
    def refraction(
        self,
        field_coordinate: FieldCoordinate | None = None,
        wavelength: float | None = None,
        sampling: SampleSize | str | int = 64,
        pupil_diameter: float | None = None,
        field_type: FieldType = "angle",
        *,
        use_higher_order_aberrations: bool = True,
    ) -> tuple[FourierPowerVectorRefraction, Any]: ...

    @abstractmethod
    def strehl_ratio(
        self,
        field_coordinate: FieldCoordinate | None = None,
        wavelength: float | None = None,
        field_type: FieldType = "angle",
        sampling: SampleSize | str | int = 128,
        psf_type: Literal["fft", "huygens"] = "huygens",
    ) -> tuple[float, Any]: ...

    @abstractmethod
    def zernike_standard_coefficients(
        self,
        field_coordinate: FieldCoordinate | None = None,
        wavelength: float | None = None,
        field_type: FieldType = "angle",
        sampling: SampleSize | str | int = 64,
        maximum_term: int = 45,
    ) -> tuple[ZernikeCoefficients, Any]: ...


class BackendSettings(TypedDict, total=False):
    """A dictionary containing the settings for the backend."""

    field_type: FieldType
    """The field type to use in the optical system. Must be one of 'angle' or 'object_height'."""

    fields: Sequence[FieldCoordinate]
    """List of field coordinates to use in the optical system."""

    wavelengths: Sequence[float]
    """List of wavelengths to use in the optical system."""

    aperture_type: ApertureType
    """The aperture type to use in the optical system. Must be one of 'float_by_stop_size', 'entrance_pupil_diameter',
    'image_f_number', or 'object_numeric_aperture'.
    """

    aperture_value: NotRequired[float]
    """The aperture value to use in the optical system. Not required for 'float_by_stop_size'."""


class BaseBackend(ABC):
    """Base class for optical simulation backends.

    Backends should implement this interface to provide a unified interface for optical simulations.
    """

    model: BaseEye | None
    settings: BackendSettings

    analysis: ClassVar[BaseAnalysisRegistry]

    @classmethod
    @abstractmethod
    def update_settings(cls, **settings: Unpack[BackendSettings]) -> None: ...

    @classmethod
    @abstractmethod
    def build_model(
        cls,
        model: EyeModel,
        *,
        start_from_index: int = 0,
        replace_existing: bool = False,
        object_distance: float = float("inf"),
        **kwargs,
    ) -> BaseEye: ...

    @classmethod
    @abstractmethod
    def clear_model(cls) -> None: ...

    @classmethod
    @abstractmethod
    def save_model(cls, filename: str | PathLike | None = None) -> None: ...

    @classmethod
    def get_setting(cls, name: str) -> Any:
        """Get a value from the backend settings.

        This method is mainly intended for internal use, to prevent the type checker from warning
        about potentially missing keys in the settings dictionary.

        Parameters
        ----------
        name : str
            The name of the setting to get.

        Returns
        -------
        Any
            The value of the setting.

        Raises
        ------
        KeyError
            If the setting does not exist.
        """
        if name not in cls.settings:
            raise KeyError(f"Setting '{name}' does not exist.")

        return cls.settings[name]

    @classmethod
    def save_settings(cls, filename: str | PathLike) -> None:
        if not str(filename).endswith(".json"):
            raise ValueError("Settings file must have a '.json' extension.")

        Path(filename).write_text(json.dumps(cls.settings, indent=4, sort_keys=True), encoding="utf-8")


class Backend(str, Enum):
    """Available backends for optical simulations."""

    OPTICSTUDIO = "opticstudio"
    OPTILAND = "optiland"


def set_backend(
    backend: Backend | Literal["opticstudio", "optiland"] = Backend.OPTICSTUDIO,
    **settings: Unpack[BackendSettings],
) -> None:
    """Set the backend to use for optical simulations.

    Parameters
    ----------
    backend : Backend | str
        The backend to use. Must be one of the values in the `Backend` enum. Defaults to `Backend.OPTICSTUDIO`.
    settings : BackendSettings, optional
        Dictionary with settings for the backend. Defaults to `None`.

    Raises
    ------
    ValueError
        If an invalid backend is specified.
    """
    global _BACKEND  # noqa: PLW0603

    if _BACKEND is not None:
        warn(
            f"The backend is already set to {_BACKEND.__name__}. "
            f"Reconfiguring the backend is not recommended and may cause issues."
        )

    if backend == Backend.OPTICSTUDIO:
        from visisipy.opticstudio import OpticStudioBackend  # noqa: PLC0415

        _BACKEND = OpticStudioBackend
        _BACKEND.initialize(**settings)
    elif backend == Backend.OPTILAND:
        from visisipy.optiland import OptilandBackend  # noqa: PLC0415

        _BACKEND = OptilandBackend
        _BACKEND.initialize(**settings)
    else:
        raise ValueError(f"Unknown backend: {backend}")


def get_backend() -> type[BaseBackend]:
    """Get the current backend.

    The backend is set to the default backend if it has not been set yet.

    Returns
    -------
    BaseBackend
        The current backend.
    """
    if _BACKEND is None:
        set_backend(_DEFAULT_BACKEND)

    return cast(type[BaseBackend], _BACKEND)


def get_oss() -> OpticStudioSystem | None:
    """Get the OpticStudioSystem instance from the current backend.

    Returns
    -------
    OpticStudioSystem | None
        The OpticStudioSystem instance if the current backend is the OpticStudio backend, otherwise `None`.
    """
    from visisipy.opticstudio import OpticStudioBackend  # noqa: PLC0415

    if _BACKEND is OpticStudioBackend:
        return OpticStudioBackend.oss

    return None


def get_optic() -> Optic | None:
    """Get the Optic instance from the current backend.

    Returns
    -------
    Optic
        The Optic instance if the current backend is the Optiland backend, otherwise `None`.
    """
    from visisipy.optiland import OptilandBackend  # noqa: PLC0415

    if _BACKEND is OptilandBackend:
        return OptilandBackend.optic

    return None


def update_settings(backend: type[BaseBackend] | None = None, **settings: Unpack[BackendSettings]):
    """Update settings on the current backend.

    Optionally, the backend can be manually specified. By default, the current backend is used.

    Parameters
    ----------
    backend : type[BaseBackend] | None
        The backend to update. If `None`, the current backend is used.
    **settings : Unpack[BackendSettings]
        The settings to update. The keys and values should match the backend's configuration schema.

    Raises
    ------
    ValueError
        If the settings are not valid for the current backend.
    """
    if backend is None:
        backend = get_backend()

    backend.update_settings(**settings)



"""Implementation of Visisipy's analyses for the OpticStudio backend."""

from __future__ import annotations

from typing import TYPE_CHECKING

from visisipy.backend import BaseAnalysisRegistry, _AnalysisMethod
from visisipy.opticstudio.analysis.cardinal_points import cardinal_points
from visisipy.opticstudio.analysis.psf import fft_psf, huygens_psf, strehl_ratio
from visisipy.opticstudio.analysis.raytrace import raytrace
from visisipy.opticstudio.analysis.refraction import refraction
from visisipy.opticstudio.analysis.zernike_coefficients import (
    zernike_standard_coefficients,
)

if TYPE_CHECKING:
    from visisipy.opticstudio.backend import OpticStudioBackend

__all__ = ("OpticStudioAnalysisRegistry",)


class OpticStudioAnalysisRegistry(BaseAnalysisRegistry):
    """Analyses for the OpticStudio backend."""

    def __init__(self, backend: OpticStudioBackend):
        super().__init__(backend)
        self._oss = backend.oss

    cardinal_points = _AnalysisMethod(cardinal_points)
    fft_psf = _AnalysisMethod(fft_psf)
    huygens_psf = _AnalysisMethod(huygens_psf)
    raytrace = _AnalysisMethod(raytrace)
    refraction = _AnalysisMethod(refraction)
    strehl_ratio = _AnalysisMethod(strehl_ratio)
    zernike_standard_coefficients = _AnalysisMethod(zernike_standard_coefficients)



"""Visisipy: VISion Simulations In PYthon."""

from __future__ import annotations

import platform
from importlib.metadata import version

from visisipy import analysis, models, optiland, plots, refraction, wavefront
from visisipy._zospy_loader import install_zospy_loader
from visisipy.backend import get_backend, set_backend, update_settings
from visisipy.models import (
    EyeGeometry,
    EyeMaterials,
    EyeModel,
    NavarroGeometry,
    NavarroMaterials,
    create_geometry,
)

__all__ = [
    "EyeGeometry",
    "EyeMaterials",
    "EyeModel",
    "NavarroGeometry",
    "NavarroMaterials",
    "analysis",
    "create_geometry",
    "get_backend",
    "models",
    "optiland",
    "plots",
    "refraction",
    "set_backend",
    "update_settings",
    "wavefront",
]

# The OpticStudio backend is only available on Windows
if platform.system() == "Windows":
    from visisipy import opticstudio

    __all__ += ["opticstudio"]

__version__ = version("visisipy")

install_zospy_loader()



"""PSF analyses for Optiland."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import numpy as np
import pandas as pd
from optiland.psf import FFTPSF, HuygensPSF

from visisipy.optiland.analysis.helpers import set_field, set_wavelength
from visisipy.types import SampleSize

if TYPE_CHECKING:
    from visisipy.optiland import OptilandBackend
    from visisipy.types import FieldCoordinate, FieldType


__all__ = ("fft_psf", "huygens_psf")


FFT_PSF_MINIMUM_PUPIL_SAMPLING = 32


def _effective_pupil_sampling(pupil_sampling: int | SampleSize) -> int:
    """Calculate the effective pupil sampling based on the given pupil sampling.

    The rationale behind this calculation is documented in the OpticStudio documentation [1]_.
    Calculating the effective pupil sampling ensures that the PSF extent is similar to OpticStudio's PSF extent.
    See also this discussion on GitHub: https://github.com/HarrisonKramer/optiland/discussions/157

    .. [1] https://ansyshelp.ansys.com/public/account/secured?returnurl=/Views/Secured/Zemax/v251/en/OpticStudio_User_Guide/OpticStudio_Help/topics/FFT_PSF.html

    Parameters
    ----------
    pupil_sampling : int
        The pupil sampling value to calculate the effective pupil sampling from.

    Returns
    -------
    int
        The effective pupil sampling value, which is the number of rays used to sample the wavefront in the exit pupil.
    """
    pupil_sampling = int(pupil_sampling)

    if pupil_sampling < FFT_PSF_MINIMUM_PUPIL_SAMPLING:
        raise ValueError("Pupil sampling must be at least 32 for FFT PSF calculation.")

    return np.floor(32 * np.sqrt(2) ** (np.log2(pupil_sampling) - 5)).astype(int)


def fft_psf(
    backend: type[OptilandBackend],
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    sampling: SampleSize | str | int = 128,
) -> tuple[pd.DataFrame, FFTPSF]:
    """Calculate the FFT Point Spread Function (PSF) at the retina surface.

    Parameters
    ----------
    backend : type[OptilandBackend]
        Reference to the Optiland backend.
    field_coordinate : tuple[float, float], optional
        The field coordinate (x, y) in mm. If `None`, the first field in Optiland is used. Defaults to `None`.
    wavelength : float, optional
        The wavelength in μm. If `None`, the first wavelength in Optiland is used. Defaults to `None`.
    field_type : Literal["angle", "object_height"], optional
        The field type. Either "angle" or "object_height". Defaults to "angle". This parameter is only used if
        `field_coordinate` is not `None`.
    sampling : SampleSize | str | int, optional
        The size of the ray grid used to sample the pupil, either string (e.g. '32x32') or int (e.g. 32). Defaults to 128.

    Returns
    -------
    DataFrame
        The PSF data as a pandas DataFrame.
    FFTPSF
        The Optiland FFTPSF object.
    """
    if not isinstance(sampling, SampleSize):
        sampling = SampleSize(sampling)

    normalized_field = set_field(backend, field_coordinate, field_type)
    wavelength = set_wavelength(backend, wavelength)

    num_rays = _effective_pupil_sampling(sampling)

    psf = FFTPSF(
        optic=backend.get_optic(),
        field=normalized_field,
        wavelength=wavelength,
        num_rays=num_rays,
        grid_size=int(2 * sampling),
    )

    (psf_extent_x, *_), (psf_extent_y, *_) = psf._get_psf_units(psf.psf)  # noqa: SLF001
    index = np.linspace(-psf_extent_x / 2, psf_extent_x / 2, psf.psf.shape[0])
    columns = np.linspace(-psf_extent_y / 2, psf_extent_y / 2, psf.psf.shape[1])

    # The PSF rows are reversed in the y-direction to match the orientation of the PSF in OpticStudio.
    df = pd.DataFrame(psf.psf[::-1, :] / 100, index=index, columns=columns)

    return df, psf


def huygens_psf(
    backend: type[OptilandBackend],
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    pupil_sampling: SampleSize | str | int = 128,
    image_sampling: SampleSize | str | int = 128,
) -> tuple[pd.DataFrame, HuygensPSF]:
    """Calculate the Huygens Point Spread Function (PSF) at the retina surface.

    Parameters
    ----------
    backend : type[OptilandBackend]
        Reference to the Optiland backend.
    field_coordinate : tuple[float, float], optional
        The field coordinate (x, y) in mm. If `None`, the first field in Optiland is used. Defaults to `None`.
    wavelength : float, optional
        The wavelength in μm. If `None`, the first wavelength in Optiland is used. Defaults to `None`.
    field_type : Literal["angle", "object_height"], optional
        The field type. Either "angle" or "object_height". Defaults to "angle". This parameter is only used if
        `field_coordinate` is not `None`.
    pupil_sampling : SampleSize | str | int, optional
        The size of the ray grid used to sample the pupil, either string (e.g. '32x32') or int (e.g. 32). Defaults to 128.
    image_sampling : SampleSize | str | int, optional
        The size of the PSF grid, either string (e.g. '32x32') or int (e.g. 32). Defaults to 128.

    Returns
    -------
    DataFrame
        The PSF data as a pandas DataFrame.
    HuygensPSF
        The Optiland HuygensPSF object.
    """
    if not isinstance(pupil_sampling, SampleSize):
        pupil_sampling = SampleSize(pupil_sampling)

    if not isinstance(image_sampling, SampleSize):
        image_sampling = SampleSize(image_sampling)

    normalized_field = set_field(backend, field_coordinate, field_type)
    wavelength = set_wavelength(backend, wavelength)

    psf = HuygensPSF(
        optic=backend.get_optic(),
        field=normalized_field,
        wavelength=wavelength,
        num_rays=int(pupil_sampling),
        image_size=int(image_sampling),
    )

    (psf_extent_x, *_), (psf_extent_y, *_) = psf._get_psf_units(psf.psf)  # noqa: SLF001
    index = np.linspace(-psf_extent_x / 2, psf_extent_x / 2, psf.psf.shape[0])
    columns = np.linspace(-psf_extent_y / 2, psf_extent_y / 2, psf.psf.shape[1])

    # The PSF rows are reversed in the y-direction to match the orientation of the PSF in OpticStudio.
    df = pd.DataFrame(psf.psf[::-1, :] / 100, index=index, columns=columns)

    return df, psf


def strehl_ratio(
    backend: type[OptilandBackend],
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    sampling: SampleSize | str | int = 128,
    psf_type: Literal["fft", "huygens"] = "huygens",
) -> tuple[float, FFTPSF | HuygensPSF]:
    """Calculate the Strehl ratio of the optical system.

    The Strehl ratio is calculated from the point spread function. Which PSF is used depends on the `psf_type` parameter.

    Parameters
    ----------
    backend : type[OptilandBackend]
        Reference to the Optiland backend.
    field_coordinate : FieldCoordinate | None
        The field coordinate at which the Strehl ratio is calculated. If `None`, the first field coordinate in
        Optiland is used.
    wavelength : float | None
        The wavelength at which the Strehl ratio is calculated. If `None`, the first wavelength in Optiland is used.
    field_type : FieldType
        The field type to be used in the analysis. Can be either "angle" or "object_height". Defaults to "angle".
        This parameter is only used when `field_coordinate` is specified.
    sampling : SampleSize | str | int
        The size of the ray grid used to sample the pupil. Can be an integer or a string in the format "NxN", where N
        is an integer. Defaults to 128.
    psf_type : Literal["fft", "huygens"]
        The type of PSF to be used for the Strehl ratio calculation. Can be either "fft" or "huygens". Defaults to "huygens".

    Returns
    -------
    float
        The Strehl ratio of the optical system at the specified field coordinate and wavelength.
    FFTPSF | HuygensPSF
        The PSF object used to calculate the Strehl ratio. The type of the object depends on the `psf_type` parameter.
    """
    if psf_type == "fft":
        _, psf = fft_psf(
            backend=backend,
            field_coordinate=field_coordinate,
            wavelength=wavelength,
            field_type=field_type,
            sampling=sampling,
        )

        return psf.strehl_ratio(), psf

    if psf_type == "huygens":
        _, psf = huygens_psf(
            backend=backend,
            field_coordinate=field_coordinate,
            wavelength=wavelength,
            field_type=field_type,
            pupil_sampling=sampling,
            image_sampling=sampling,
        )

        return psf.strehl_ratio(), psf

    raise NotImplementedError(f"PSF type '{psf_type}' is not implemented.")



"""Build and manage eye models in OpticStudio."""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from visisipy.models import BaseEye, EyeModel
from visisipy.opticstudio.surfaces import OpticStudioSurface, make_surface

if TYPE_CHECKING:
    from zospy.zpcore import OpticStudioSystem


__all__ = ("OpticStudioEye", "OpticStudioReverseEye")


class BaseOpticStudioEye(BaseEye):
    """Base class for OpticStudio eye models."""

    @abstractmethod
    def __init__(self, model: EyeModel): ...

    @property
    @abstractmethod
    def pupil(self) -> OpticStudioSurface:
        """Iris / pupil surface."""
        ...

    @abstractmethod
    def build(
        self,
        oss: OpticStudioSystem,
        *,
        start_from_index: int = 0,
        replace_existing: bool = False,
        object_distance: float = float("inf"),
    ):
        """Create the eye in OpticStudio.

        Create the eye model in the provided `OpticStudioSystem` `oss`, starting from `start_from_index`.
        The iris (pupil) is located at the STOP surface, and the retina at the IMAGE surface. For the other
        parts, new surfaces will be inserted by default. If `replace_existing` is set to `True`, existing
        surfaces will be overwritten.

        Parameters
        ----------
        oss : zospy.zpcore.OpticStudioSystem
            OpticStudioSystem in which the eye model is created.
        start_from_index : int
            Index at which the first  surface of the eye is located.
        replace_existing : bool
            If `True`, replaces existing surfaces instead of inserting new ones. Defaults to `False`.

        Raises
        ------
        AssertionError
            If the retina is not located at the IMAGE surface.
        """
        ...

    @property
    @abstractmethod
    def eye_model(self) -> EyeModel: ...

    @property
    def surfaces(self) -> dict[str, OpticStudioSurface]:
        """Dictionary with surface names as keys and surfaces as values."""
        return {k.lstrip("_"): v for k, v in self.__dict__.items() if isinstance(v, OpticStudioSurface)}

    def relink_surfaces(self, oss: OpticStudioSystem) -> bool:
        """Link surfaces to OpticStudio surfaces based on their comments.

        Attempt to re-link the surfaces of the eye to surfaces defined in OpticStudio, using the surface's comments.

        Parameters
        ----------
        oss : zospy.zpcore.OpticStudioSystem

        Returns
        -------
        bool
            `True` if all surfaces could be relinked, `False` otherwise.
        """
        result = [s.relink_surface(oss) for s in self.surfaces.values()]

        return all(result)


class OpticStudioEye(BaseOpticStudioEye):
    """Eye model in OpticStudio."""

    def __init__(self, eye_model: EyeModel) -> None:
        """Create a new OpticStudio eye model.

        Parameters
        ----------
        eye_model : EyeModel
            Eye model specification from which the OpticStudio eye model is created.
        """
        self._eye_model = eye_model

        self._cornea_front = make_surface(eye_model.geometry.cornea_front, eye_model.materials.cornea, "cornea front")
        self._cornea_back = make_surface(
            eye_model.geometry.cornea_back,
            eye_model.materials.aqueous,
            "cornea back / aqueous",
        )
        self._pupil = make_surface(eye_model.geometry.pupil, eye_model.materials.aqueous, "pupil")
        self._lens_front = make_surface(eye_model.geometry.lens_front, eye_model.materials.lens, "lens front")
        self._lens_back = make_surface(
            eye_model.geometry.lens_back,
            eye_model.materials.vitreous,
            "lens back / vitreous",
        )
        self._retina = make_surface(eye_model.geometry.retina, eye_model.materials.vitreous, "retina")

    @property
    def eye_model(self) -> EyeModel:
        """Eye model specification from which the OpticStudio eye model is created."""
        return self._eye_model

    @property
    def cornea_front(self) -> OpticStudioSurface:
        """Cornea front surface."""
        return self._cornea_front

    @cornea_front.setter
    def cornea_front(self, value: OpticStudioSurface) -> None:
        self._cornea_front = value

    @property
    def cornea_back(self) -> OpticStudioSurface:
        """Cornea back surface."""
        return self._cornea_back

    @cornea_back.setter
    def cornea_back(self, value: OpticStudioSurface) -> None:
        self._cornea_back = value

    @property
    def pupil(self) -> OpticStudioSurface:
        """Iris / pupil surface."""
        return self._pupil

    @pupil.setter
    def pupil(self, value: OpticStudioSurface) -> None:
        self._pupil = value

    @property
    def lens_front(self) -> OpticStudioSurface:
        """Lens front surface."""
        return self._lens_front

    @lens_front.setter
    def lens_front(self, value: OpticStudioSurface) -> None:
        self._lens_front = value

    @property
    def lens_back(self) -> OpticStudioSurface:
        """Lens back surface."""
        return self._lens_back

    @lens_back.setter
    def lens_back(self, value: OpticStudioSurface) -> None:
        self._lens_back = value

    @property
    def retina(self) -> OpticStudioSurface:
        """Retina surface."""
        return self._retina

    @retina.setter
    def retina(self, value: OpticStudioSurface) -> None:
        self._retina = value

    def build(
        self,
        oss: OpticStudioSystem,
        *,
        start_from_index: int = 0,
        replace_existing: bool = False,
        object_distance: float = float("inf"),
    ):
        """Create the eye in OpticStudio.

        Create the eye model in the provided `OpticStudioSystem` `oss`, starting from `start_from_index`.
        The iris (pupil) is located at the STOP surface, and the retina at the IMAGE surface. For the other
        parts, new surfaces will be inserted by default. If `replace_existing` is set to `True`, existing
        surfaces will be overwritten.

        Parameters
        ----------
        oss : zospy.zpcore.OpticStudioSystem
            OpticStudioSystem in which the eye model is created.
        start_from_index : int
            Index of the surface after which the eye model will be built. Because the pupil will be located at the stop surface,
            `start_from_index` must be smaller than the index of the stop surface.
        replace_existing : bool
            If `True`, replaces existing surfaces instead of inserting new ones. Defaults to `False`.
        object_distance : float, optional
            Distance from the object surface (or the surface before the eye model) to the eye model. Defaults to infinity.

        Raises
        ------
        ValueError
            If the pupil is not located at the stop position.
            If the retina is not located at the image surface.
        """
        if start_from_index >= oss.LDE.StopSurface:
            message = "'start_from_index' must be smaller than the index of the stop surface."
            raise ValueError(message)

        if object_distance != float("inf"):
            oss.LDE.GetSurfaceAt(start_from_index).Thickness = object_distance

        cornea_front_index = self.cornea_front.build(
            oss, position=start_from_index + 1, replace_existing=replace_existing
        )
        cornea_back_index = self.cornea_back.build(
            oss, position=cornea_front_index + 1, replace_existing=replace_existing
        )
        pupil_index = self.pupil.build(oss, position=cornea_back_index + 1, replace_existing=True)
        lens_front_index = self.lens_front.build(oss, position=pupil_index + 1, replace_existing=replace_existing)
        lens_back_index = self.lens_back.build(oss, position=lens_front_index + 1, replace_existing=replace_existing)
        self.retina.build(oss, position=lens_back_index + 1, replace_existing=True)

        # Sanity checks
        if not self.pupil.surface.IsStop:
            message = "The pupil is not located at the stop position."
            raise ValueError(message)

        if not self.retina.surface.IsImage:
            message = "The retina is not located at the image position."
            raise ValueError(message)


class OpticStudioReverseEye(BaseOpticStudioEye):  # pragma: no cover
    def __init__(self, eye_model: EyeModel):
        self._eye_model = eye_model

        self._retina = OpticStudioSurface(
            comment="retina / vitreous",
            radius=-1 * eye_model.geometry.retina_curvature,
            conic=eye_model.geometry.retina_asphericity,
            thickness=eye_model.geometry.vitreous_thickness,
            refractive_index=eye_model.materials.vitreous_refractive_index,
        )
        self._lens_back = OpticStudioSurface(
            comment="lens back",
            radius=-1 * eye_model.geometry.lens_back_curvature,
            conic=eye_model.geometry.lens_back_asphericity,
            thickness=eye_model.geometry.lens_thickness,
            refractive_index=eye_model.materials.lens_refractive_index,
        )
        self._lens_front = OpticStudioSurface(
            comment="lens front",
            radius=-1 * eye_model.geometry.lens_front_curvature,
            conic=-eye_model.geometry.lens_front_asphericity,
            refractive_index=eye_model.materials.aqueous_refractive_index,
        )
        self._iris = OpticStudioSurface(
            comment="iris",
            is_stop=True,
            refractive_index=eye_model.materials.aqueous_refractive_index,
            semi_diameter=eye_model.geometry.iris_radius,
        )
        self._aqueous = OpticStudioSurface(
            comment="aqueous",
            thickness=eye_model.geometry.anterior_chamber_depth,
            refractive_index=eye_model.materials.aqueous_refractive_index,
        )
        self._cornea_back = OpticStudioSurface(
            comment="cornea back",
            radius=-1 * eye_model.geometry.cornea_back_curvature,
            conic=eye_model.geometry.cornea_back_asphericity,
            thickness=eye_model.geometry.cornea_thickness,
            refractive_index=eye_model.materials.cornea_refractive_index,
        )
        self._cornea_front = OpticStudioSurface(
            comment="cornea front",
            radius=-1 * eye_model.geometry.cornea_front_curvature,
            conic=eye_model.geometry.cornea_front_asphericity,
        )

    @property
    def eye_model(self) -> EyeModel:
        return self._eye_model

    @property
    def cornea_front(self) -> OpticStudioSurface:
        return self._cornea_front

    @cornea_front.setter
    def cornea_front(self, value: OpticStudioSurface) -> None:
        self._cornea_front = value

    @property
    def cornea_back(self) -> OpticStudioSurface:
        return self._cornea_back

    @cornea_back.setter
    def cornea_back(self, value: OpticStudioSurface) -> None:
        self._cornea_back = value

    @property
    def aqueous(self) -> OpticStudioSurface:
        return self._aqueous

    @aqueous.setter
    def aqueous(self, value: OpticStudioSurface) -> None:
        self._aqueous = value

    @property
    def iris(self) -> OpticStudioSurface:
        return self._iris

    @iris.setter
    def iris(self, value: OpticStudioSurface) -> None:
        self._iris = value

    @property
    def lens_front(self) -> OpticStudioSurface:
        return self._lens_front

    @lens_front.setter
    def lens_front(self, value: OpticStudioSurface) -> None:
        self._lens_front = value

    @property
    def lens_back(self) -> OpticStudioSurface:
        return self._lens_back

    @lens_back.setter
    def lens_back(self, value: OpticStudioSurface) -> None:
        self._lens_back = value

    @property
    def retina(self) -> OpticStudioSurface:
        return self._retina

    @retina.setter
    def retina(self, value: OpticStudioSurface) -> None:
        self._retina = value

    def build(
        self,
        oss: OpticStudioSystem,
        *,
        start_from_index: int = 0,
        replace_existing: bool = False,
    ):
        self.retina.build(oss, position=start_from_index, replace_existing=True)
        self.lens_back.build(oss, position=start_from_index + 1, replace_existing=replace_existing)
        self.lens_front.build(oss, position=start_from_index + 2, replace_existing=replace_existing)
        self.iris.build(oss, position=start_from_index + 3, replace_existing=True)
        self.aqueous.build(oss, position=start_from_index + 4, replace_existing=replace_existing)
        self.cornea_back.build(oss, position=start_from_index + 5, replace_existing=replace_existing)
        self.cornea_front.build(oss, position=start_from_index + 6, replace_existing=replace_existing)

        # Sanity checks
        if not self.retina.surface.IsObject:
            message = "The retina is not located at the object position."
            raise ValueError(message)



"""PSF analyses for OpticStudio."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import zospy as zp

from visisipy.opticstudio.analysis.helpers import set_field, set_wavelength
from visisipy.types import FieldCoordinate, FieldType, SampleSize

if TYPE_CHECKING:
    import pandas as pd
    from zospy.analyses.psf.huygens_psf import HuygensPSFResult

    from visisipy.opticstudio import OpticStudioBackend


def fft_psf(
    backend: type[OpticStudioBackend],
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    sampling: SampleSize | str | int = 128,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Calculate the FFT Point Spread Function (PSF) at the retina surface.

    Parameters
    ----------
    backend : type[OpticStudioBackend]
        Reference to the OpticStudio backend.
    field_coordinate : tuple[float, float], optional
        The field coordinate (x, y) in mm. If `None`, the first field in OpticStudio is used. Defaults to `None`.
    wavelength : float, optional
        The wavelength in μm. If `None`, the first wavelength in OpticStudio is used. Defaults to `None`.
    field_type : Literal["angle", "object_height"], optional
        The field type. Either "angle" or "object_height". Defaults to "angle". This parameter is only used if
        `field_coordinate` is not `None`.
    sampling : SampleSize | str | int, optional
        The size of the ray grid used to sample the pupil, either string (e.g. '32x32') or int (e.g. 32). Defaults to 128.

    Returns
    -------
    DataFrame
        The PSF data as a pandas DataFrame.
    """

    if not isinstance(sampling, SampleSize):
        sampling = SampleSize(sampling)

    wavelength_number = set_wavelength(backend, wavelength)
    field_number = set_field(backend, field_coordinate, field_type)

    psf_result = zp.analyses.psf.FFTPSF(
        sampling=str(sampling),
        display=str(2 * sampling),
        wavelength=wavelength_number,
        field=field_number,
        psf_type=zp.constants.Analysis.Settings.Psf.FftPsfType.Linear,
        surface="Image",
        normalize=False,
    ).run(backend.get_oss())

    if psf_result.data is None:
        raise ValueError("Failed to run FFT PSF analysis.")

    return psf_result.data, psf_result.data


def huygens_psf(
    backend: type[OpticStudioBackend],
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    pupil_sampling: SampleSize | str | int = 128,
    image_sampling: SampleSize | str | int = 128,
) -> tuple[pd.DataFrame, HuygensPSFResult]:
    """Calculate the Huygens Point Spread Function (PSF) at the retina surface.

    Parameters
    ----------
    backend : type[OpticStudioBackend]
        Reference to the OpticStudio backend.
    field_coordinate : tuple[float, float], optional
        The field coordinate (x, y) in mm. If `None`, the first field in OpticStudio is used. Defaults to `None`.
    wavelength : float, optional
        The wavelength in μm. If `None`, the first wavelength in OpticStudio is used. Defaults to `None`.
    field_type : Literal["angle", "object_height"], optional
        The field type. Either "angle" or "object_height". Defaults to "angle". This parameter is only used if
        `field_coordinate` is not `None`.
    pupil_sampling : SampleSize | str | int, optional
        The size of the ray grid used to sample the pupil, either string (e.g. '32x32') or int (e.g. 32). Defaults to 128.
    image_sampling : SampleSize | str | int, optional
        The size of the PSF grid, either string (e.g. '32x32') or int (e.g. 32). Defaults to 128.

    Returns
    -------
    DataFrame
        The PSF data as a pandas DataFrame.
    HuygensPSFData
        The Huygens PSF result from OpticStudio.
    """

    if not isinstance(pupil_sampling, SampleSize):
        pupil_sampling = SampleSize(pupil_sampling)

    if not isinstance(image_sampling, SampleSize):
        image_sampling = SampleSize(image_sampling)

    wavelength_number = set_wavelength(backend, wavelength)
    field_number = set_field(backend, field_coordinate, field_type)

    psf_result = zp.analyses.psf.HuygensPSFAndStrehlRatio(
        pupil_sampling=str(pupil_sampling),
        image_sampling=str(image_sampling),
        image_delta=0,
        rotation=0,
        wavelength=wavelength_number,
        field=field_number,
        psf_type="Linear",
        show_as="Surface",
        use_polarization=False,
        use_centroid=False,
        normalize=False,
    ).run(backend.get_oss())

    return psf_result.data.psf, psf_result.data


def strehl_ratio(
    backend: type[OpticStudioBackend],
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    sampling: SampleSize | str | int = 128,
    psf_type: Literal["fft", "huygens"] = "huygens",
) -> tuple[float, HuygensPSFResult]:
    """Calculate the Strehl ratio of the optical system.

    The Strehl ratio is calculated from the point spread function. Which PSF is used depends on the `psf_type` parameter.

    Parameters
    ----------
    backend : type[OpticStudioBackend]
        Reference to the OpticStudio backend.
    field_coordinate : FieldCoordinate | None
        The field coordinate at which the Strehl ratio is calculated. If `None`, the first field coordinate in
        OpticStudio is used.
    wavelength : float | None
        The wavelength at which the Strehl ratio is calculated. If `None`, the first wavelength in OpticStudio is used.
    field_type : FieldType
        The field type to be used in the analysis. Can be either "angle" or "object_height". Defaults to "angle".
        This parameter is only used when `field_coordinate` is specified.
    sampling : SampleSize | str | int
        The size of the ray grid used to sample the pupil. Can be an integer or a string in the format "NxN", where N
        is an integer. Defaults to 128.
    psf_type : Literal["fft", "huygens"]
        The type of PSF to be used for the Strehl ratio calculation. Can be either "fft" or "huygens". Defaults to "huygens";
        OpticStudio's FFT PSF does not support calculating the Strehl ratio, so only "huygens" is supported.

    Returns
    -------
    float
        The Strehl ratio of the optical system at the specified field coordinate and wavelength.
    HuygensPSFResult
        The PSF object used to calculate the Strehl ratio. The type of the object depends on the `psf_type` parameter.
    """
    if psf_type == "fft":
        raise NotImplementedError("OpticStudio does not support obtaining the Strehl ratio from the FFT PSF.")

    if psf_type == "huygens":
        _, psf = huygens_psf(
            backend=backend,
            field_coordinate=field_coordinate,
            wavelength=wavelength,
            field_type=field_type,
            pupil_sampling=sampling,
            image_sampling=sampling,
        )

        return psf.strehl_ratio, psf

    raise NotImplementedError(f"PSF type '{psf_type}' is not implemented. Only 'huygens' is supported.")



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from visisipy.optiland import OptilandEye

if TYPE_CHECKING:
    from optiland.optic import Optic

    from visisipy import EyeModel


class TestOptilandEye:
    def test_init(self, eye_model):
        opticstudio_eye = OptilandEye(eye_model)

        assert opticstudio_eye.eye_model == eye_model

    def test_build_cornea_front(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic)

        assert optic.surface_group.get_thickness(1) == pytest.approx(eye_model.geometry.cornea_front.thickness)
        assert optic.surface_group.radii[1] == pytest.approx(eye_model.geometry.cornea_front.radius)
        assert optic.surface_group.conic[1] == pytest.approx(eye_model.geometry.cornea_front.asphericity)

    def test_build_cornea_back(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic)

        assert optic.surface_group.get_thickness(2) == pytest.approx(eye_model.geometry.cornea_back.thickness)
        assert optic.surface_group.radii[2] == pytest.approx(eye_model.geometry.cornea_back.radius)
        assert optic.surface_group.conic[2] == pytest.approx(eye_model.geometry.cornea_back.asphericity)

    def test_build_pupil(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic)

        # assert optic.surface_group.surfaces[3].semi_aperture
        assert optic.surface_group.stop_index == 3
        assert optic.surface_group.get_thickness(3) == 0
        assert optic.surface_group.radii[3] == float("inf")

    def test_build_lens_front(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic)

        assert optic.surface_group.get_thickness(4) == pytest.approx(eye_model.geometry.lens_front.thickness)
        assert optic.surface_group.radii[4] == pytest.approx(eye_model.geometry.lens_front.radius)
        assert optic.surface_group.conic[4] == pytest.approx(eye_model.geometry.lens_front.asphericity)

    def test_build_lens_back(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic)

        assert optic.surface_group.get_thickness(5) == pytest.approx(eye_model.geometry.lens_back.thickness)
        assert optic.surface_group.radii[5] == pytest.approx(eye_model.geometry.lens_back.radius)
        assert optic.surface_group.conic[5] == pytest.approx(eye_model.geometry.lens_back.asphericity)

    def test_build_retina(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic)

        assert optic.surface_group.radii[6] == pytest.approx(eye_model.geometry.retina.radius)
        assert optic.surface_group.conic[6] == pytest.approx(eye_model.geometry.retina.asphericity)
        assert optic.image_surface == optiland_eye.retina.surface

    def test_build_pupil_not_stop_raises_valueerror(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.lens_front._is_stop = True

        with pytest.raises(ValueError, match="The pupil is not located at the stop position"):
            optiland_eye.build(optic)

    def test_build_retina_not_image_raises_valueerror(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)

        # Surfaces are inserted before an existing surface if a surface is already present at the specified index.
        optic.add_surface(index=0, comment="Dummy 1")
        optic.add_surface(index=0, comment="Dummy 2")

        with pytest.raises(ValueError, match="The retina is not located at the image position"):
            optiland_eye.build(optic)

    def test_build_start_from_index(self, optic: Optic, eye_model: EyeModel):
        optic.add_surface(index=0, comment="Dummy 1")
        optic.add_surface(index=1, comment="Dummy 1")

        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic, start_from_index=1)

        assert optic.surface_group.surfaces[2].comment == "cornea front"
        assert optic.surface_group.get_thickness(2) == eye_model.geometry.cornea_front.thickness

    @pytest.mark.parametrize("n_surfaces,index", [(0, 1), (0, 3), (2, 2), (2, 3)])
    def test_build_start_from_index_invalid(self, n_surfaces: int, index: int, optic: Optic, eye_model: EyeModel):
        for i in range(n_surfaces):
            optic.add_surface(index=i, comment=f"Dummy {i}")

        optiland_eye = OptilandEye(eye_model)

        with pytest.raises(
            ValueError, match="'start_from_index' can be at most the index of the last surface in the system"
        ):
            optiland_eye.build(optic, start_from_index=index)

    def test_build_object_distance(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic, object_distance=10)

        assert optic.surface_group.get_thickness(0) == 10

    def test_build_object_distance_nonempty_system(self, optic: Optic, eye_model: EyeModel):
        optic.add_surface(index=0, comment="Object")

        optiland_eye = OptilandEye(eye_model)

        with pytest.raises(ValueError, match="Cannot set a custom object distance if the optical system is not empty"):
            optiland_eye.build(optic, object_distance=10)

    def test_get_cornea_front_comment(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic)

        assert optiland_eye.cornea_front.comment == "cornea front"

    def test_get_cornea_back_comment(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic)

        assert optiland_eye.cornea_back.comment == "cornea back / aqueous"

    def test_get_pupil_comment(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic)

        assert optiland_eye.pupil.comment == "pupil"

    def test_get_lens_front_comment(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic)

        assert optiland_eye.lens_front.comment == "lens front"

    def test_get_lens_back_comment(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic)

        assert optiland_eye.lens_back.comment == "lens back / vitreous"

    def test_get_retina_comment(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic)

        assert optiland_eye.retina.comment == "retina"

    @pytest.mark.parametrize("build", [True, False])
    def test_surfaces(self, build: bool, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)

        if build:
            optiland_eye.build(optic)

        assert optiland_eye.surfaces == {
            "cornea_front": optiland_eye.cornea_front,
            "cornea_back": optiland_eye.cornea_back,
            "pupil": optiland_eye.pupil,
            "lens_front": optiland_eye.lens_front,
            "lens_back": optiland_eye.lens_back,
            "retina": optiland_eye.retina,
        }

    def test_update_surfaces(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic)

        optiland_eye.update_surfaces("comment", "new comment")

        assert optiland_eye.cornea_front.comment == "new comment"
        assert optiland_eye.cornea_back.comment == "new comment"
        assert optiland_eye.pupil.comment == "new comment"
        assert optiland_eye.lens_front.comment == "new comment"
        assert optiland_eye.lens_back.comment == "new comment"
        assert optiland_eye.retina.comment == "new comment"

    def test_update_surfaces_subset(self, optic: Optic, eye_model: EyeModel):
        optiland_eye = OptilandEye(eye_model)
        optiland_eye.build(optic)

        optiland_eye.update_surfaces("comment", "new comment", surface_names=["cornea_front", "lens_back"])

        assert optiland_eye.cornea_front.comment == "new comment"
        assert optiland_eye.cornea_back.comment != "new comment"
        assert optiland_eye.pupil.comment != "new comment"
        assert optiland_eye.lens_front.comment != "new comment"
        assert optiland_eye.lens_back.comment == "new comment"
        assert optiland_eye.retina.comment != "new comment"


def test_model_with_no_surface(optic, three_surface_eye_model):
    optiland_eye = OptilandEye(three_surface_eye_model)
    optiland_eye.build(optic)

    assert optic.surface_group.num_surfaces == 6
    assert optiland_eye.cornea_front.surface == optic.surface_group.surfaces[1]
    assert optiland_eye.pupil.surface == optic.surface_group.surfaces[2]
    assert optiland_eye.lens_front.surface == optic.surface_group.surfaces[3]
    assert optiland_eye.lens_back.surface == optic.surface_group.surfaces[4]
    assert optiland_eye.retina.surface == optic.surface_group.surfaces[5]
    assert optiland_eye.cornea_back.surface is None



"""Models for the ocular geometry."""

from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, field
from typing import Any, NamedTuple, TypeVar
from warnings import warn

import numpy as np

from visisipy.types import TypedDict, Unpack
from visisipy.wavefront import ZernikeCoefficients

__all__ = (
    "EyeGeometry",
    "NavarroGeometry",
    "NoSurface",
    "StandardSurface",
    "Stop",
    "Surface",
    "create_geometry",
)


@dataclass
class Surface(ABC):  # noqa: B024
    """Base class for optical surfaces.

    Attributes
    ----------
    thickness : float
        The thickness of the surface. Default is 0.
    """

    thickness: float = 0


class _EllipsoidRadii(NamedTuple):
    z: float
    y: float
    x: float


@dataclass
class StandardSurface(Surface):
    """Standard conic surface.

    Represents surfaces as conic sections (spheres, ellipsoids, paraboloids, hyperboloids), defined in terms of their
    radius of curvature and asphericity. For the asphericity, the following definition is used:

    .. math::
        k = - \varepsilon^2 = - \\left( 1 - \frac{b^2}{a^2} \right)

    with :math:`k` the asphericity, :math:`\varepsilon` the eccentricity, :math:`a` the ellipsoid axis parallel to
    the optical axis and :math:`b` the ellipsoid axis perpendicular to the optical axis. This is the definition used
    in OpticStudio. Note that the meaning of :math:`a` and :math:`b` differs from the standard definition of an ellipse,
    where they are the semi-major and semi-minor axes.

    Attributes
    ----------
    radius : float
        The radius of the surface. Default is infinity.
    asphericity : float
        The asphericity of the surface. Default is 0.
    thickness : float
        The thickness of the surface. Default is 0.
    semi_diameter : float | None
        The semi-diameter of the surface aperture. Default is `None`.

    Methods
    -------
    half_axes(self) -> tuple[float, float]:
        Calculates and returns the half axes of the surface.
    """

    radius: float = float("inf")
    asphericity: float = 0
    thickness: float = 0
    semi_diameter: float | None = None
    is_stop: bool = False

    @property
    def ellipsoid_radii(self) -> _EllipsoidRadii:
        """Calculates and returns the ellipsoid radii (semi-axes) of the surface.

        This works only if the surface is an ellipsoid (asphericity > -1), otherwise a NotImplementedError is raised.
        A tuple of the radii along the z, y and x axes is returned, where the z axis is the optical axis.

        Returns
        -------
        tuple[float, float, float]
            The ellipsoid radii (semi-axes) of the surface.

        Raises
        ------
        NotImplementedError
            If the surface is not an ellipsoid (asphericity <= -1).
        """
        if self.asphericity > -1:
            axial = self.radius / (self.asphericity + 1)
            radial = abs(self.radius / np.sqrt(self.asphericity + 1))
            return _EllipsoidRadii(
                z=axial,
                y=radial,
                x=radial,
            )

        raise NotImplementedError(
            f"Half axes are only defined for ellipsoids (asphericity > -1), got {self.asphericity=}"
        )


@dataclass
class Stop(StandardSurface):
    """Stop surface.

    This surface represents the aperture stop of an optical system.

    Attributes
    ----------
    semi_diameter : float
        The semi-diameter of the aperture stop. Default is 1.
    thickness : float
        The thickness of the stop. Default is 0.
    """

    semi_diameter: float = 1.0
    thickness: float = 0
    is_stop: bool = True


@dataclass
class ZernikeStandardSagSurface(StandardSurface):
    """Zernike standard coefficients surface with surface deformations.

    Represents a surface with surface deformations described by Zernike polynomials.

    Attributes
    ----------
    zernike_coefficients : ZernikeCoefficients | dict[int, float]
        The Zernike coefficients of the surface. Default is an empty dictionary.
    extrapolate : bool
        If True, the Zernike coefficients will be considered even if the ray lands beyond the normalization radius.
        Default is `True`.
    zernike_decenter_x : float
        Decentration of the Zernike terms with respect to the conical and aspherical terms in the x-direction.
        Default is 0.
    zernike_decenter_y : float
        Decentration of the Zernike terms with respect to the conical and aspherical terms in the y-direction.
        Default is 0.
    maximum_term : int
        The maximum Zernike term to consider. Default is 0.
    norm_radius : float
        The normalization radius for the Zernike coefficients, in lens units (usually mm). Default is 100.

    Raises
    ------
    ValueError
        If the Zernike coefficients contain terms that are greater than the maximum term.
    """

    zernike_coefficients: ZernikeCoefficients | dict[int, float] = field(default_factory=dict)
    extrapolate: bool = True
    zernike_decenter_x: float = 0
    zernike_decenter_y: float = 0
    maximum_term: int = 0
    norm_radius: float = 100

    def __post_init__(self):
        if any(key > self.maximum_term for key in self.zernike_coefficients):
            raise ValueError("The Zernike coefficients contain terms that are greater than the maximum term.")

        self.zernike_coefficients = ZernikeCoefficients(self.zernike_coefficients)


@dataclass
class ZernikeStandardPhaseSurface(StandardSurface):
    """Zernike standard coefficients surface with wavefront aberrations.

    Represents a surface with wavefront aberrations described by Zernike polynomials.

    Attributes
    ----------
    zernike_coefficients : ZernikeCoefficients | dict[int, float]
        The Zernike coefficients of the surface. Default is an empty dictionary.
    extrapolate : bool
        If True, the Zernike coefficients will be considered even if the ray lands beyond the normalization radius.
        Default is `True`.
    diffraction_order : float
        The diffraction order of the surface. Default is 0.
    maximum_term : int
        The maximum Zernike term to consider. Default is 0.
    norm_radius : float
        The normalization radius for the Zernike coefficients, in lens units (usually mm). Default is 100.

    Raises
    ------
    ValueError
        If the Zernike coefficients contain terms that are greater than the maximum term.
    """

    zernike_coefficients: ZernikeCoefficients | dict[int, float] = field(default_factory=dict)
    extrapolate: bool = True
    diffraction_order: float = 1
    maximum_term: int = 0
    norm_radius: float = 100

    def __post_init__(self):
        if any(key > self.maximum_term for key in self.zernike_coefficients):
            raise ValueError("The Zernike coefficients contain terms that are greater than the maximum term.")

        self.zernike_coefficients = ZernikeCoefficients(self.zernike_coefficients)


@dataclass
class NoSurface(Surface):
    """A surface that does not exist.

    This surface is used to indicate that a surface is not present in the optical system.
    It can be used to define three-surface schematic eyes and reduced eye models.

    .. note::
       This surface does not modify the optical system, i.e. surfaces of the `NoSurface` type are not built
       when calling `EyeModel.build`. This means the properties of the preceding surface, e.g. the refractive index,
       will propagate to the next surface.
    """

    # Thickness is present on all surface instances, but cannot be set from the constructor.
    thickness: int = field(default=0, init=False)


class EyeGeometry:
    """Geometric parameters of an eye.

    Sizes are specified in mm. This class is mainly intended as a base class for more specific eye models.
    When used directly, all surfaces need to be specified manually.

    Attributes
    ----------
    cornea_front : Surface
        The front surface of the cornea.
    cornea_back : Surface
        The back surface of the cornea.
    pupil : Surface
        The pupil of the eye.
    lens_front : Surface
        The front surface of the lens.
    lens_back : Surface
        The back surface of the lens.
    retina : Surface
        The retina of the eye.

    Methods
    -------
    axial_length(self) -> float:
        Calculates and returns the axial length of the eye.
    cornea_thickness(self) -> float:
        Returns the thickness of the cornea.
    anterior_chamber_depth(self) -> float:
        Returns the depth of the anterior chamber.
    lens_thickness(self) -> float:
        Returns the thickness of the lens.
    vitreous_thickness(self) -> float:
        Returns the thickness of the vitreous.
    """

    def __init__(
        self,
        cornea_front: StandardSurface | None = None,
        cornea_back: StandardSurface | None = None,
        pupil: Stop | None = None,
        lens_front: StandardSurface | None = None,
        lens_back: StandardSurface | None = None,
        retina: StandardSurface | None = None,
    ):
        """Initializes an instance of the EyeGeometry class.

        This creates an instance of the EyeGeometry class with the specified surfaces. If a surface is not provided,
        the default values of `StandardSurface` and `Stop` will be used. This class is mainly intended as a base
        class, and instantiating it without specifying the surfaces will not result in a realistic eye model. To
        initialize a realistic eye model, use a subclass such as `NavarroGeometry`.

        Parameters
        ----------
        cornea_front : Surface, optional
            The front surface of the cornea. If not provided, a default Surface instance will be used.
        cornea_back : Surface, optional
            The back surface of the cornea. If not provided, a default Surface instance will be used.
        pupil : Surface, optional
            The pupil of the eye. If not provided, a default Stop instance will be used.
        lens_front : Surface, optional
            The front surface of the lens. If not provided, a default Surface instance will be used.
        lens_back : Surface, optional
            The back surface of the lens. If not provided, a default Surface instance will be used.
        retina : StandardSurface, optional
            The retina of the eye. If not provided, a default Surface instance will be used.
        """

        self.cornea_front = cornea_front or StandardSurface()
        self.cornea_back = cornea_back or StandardSurface()
        self.pupil = pupil or Stop()
        self.lens_front = lens_front or StandardSurface()
        self.lens_back = lens_back or StandardSurface()
        self.retina = retina or StandardSurface()

    @property
    def cornea_front(self) -> StandardSurface:
        """Cornea front geometry."""
        return self._cornea_front

    @cornea_front.setter
    def cornea_front(self, value: StandardSurface) -> None:
        self._cornea_front = value

    @property
    def cornea_back(self) -> StandardSurface:
        """Cornea back geometry."""
        return self._cornea_back

    @cornea_back.setter
    def cornea_back(self, value: StandardSurface) -> None:
        self._cornea_back = value

    @property
    def pupil(self) -> Stop:
        """Pupil geometry."""
        return self._pupil

    @pupil.setter
    def pupil(self, value: Stop) -> None:
        if not value.is_stop:
            raise ValueError("The pupil surface must be a stop.")

        self._pupil = value

    @property
    def lens_front(self) -> StandardSurface:
        """Lens front geometry."""
        return self._lens_front

    @lens_front.setter
    def lens_front(self, value: StandardSurface) -> None:
        self._lens_front = value

    @property
    def lens_back(self) -> StandardSurface:
        """Lens back geometry."""
        return self._lens_back

    @lens_back.setter
    def lens_back(self, value: StandardSurface) -> None:
        self._lens_back = value

    @property
    def retina(self) -> StandardSurface:
        """Retina geometry."""
        return self._retina

    @retina.setter
    def retina(self, value: StandardSurface) -> None:
        if value.asphericity <= -1:
            raise ValueError(f"Only an elliptical retina is allowed (asphericity > -1), got {value.asphericity=}")

        self._retina = value

    @property
    def axial_length(self) -> float:
        """Axial length of the eye, in mm."""
        return (
            self.cornea_front.thickness
            + self.cornea_back.thickness
            + self.pupil.thickness
            + self.lens_front.thickness
            + self.lens_back.thickness
        )

    @property
    def cornea_thickness(self) -> float:
        """Thickness of the cornea, in mm."""
        return self.cornea_front.thickness

    @property
    def anterior_chamber_depth(self) -> float:
        """Depth of the anterior chamber, in mm."""
        return self.cornea_back.thickness

    @property
    def pupil_lens_distance(self) -> float:
        """Distance between the pupil and the lens, in mm."""
        return self.pupil.thickness

    @property
    def lens_thickness(self) -> float:
        """Thickness of the crystalline lens, in mm."""
        return self.lens_front.thickness

    @property
    def vitreous_thickness(self) -> float:
        """Thickness of the vitreous body, in mm."""
        return self.lens_back.thickness

    def reverse(self):
        raise NotImplementedError


class NavarroGeometry(EyeGeometry):
    """Geometric parameters of the Navarro schematic eye.

    This schematic eye is based on the Navarro model as described in [1]_.
    Sizes are specified in mm.

     .. [1] Escudero-Sanz, I., & Navarro, R. (1999). Off-axis aberrations of a wide-angle schematic eye model.
            JOSA A, 16(8), 1881-1891. https://doi.org/10.1364/JOSAA.16.001881

    Attributes
    ----------
    cornea_front : StandardSurface
        The front surface of the cornea.
    cornea_back : StandardSurface
        The back surface of the cornea.
    pupil : Stop
        The pupil of the eye.
    lens_front : StandardSurface
        The front surface of the lens.
    lens_back : StandardSurface
        The back surface of the lens.
    retina : StandardSurface
        The retina of the eye.

    Examples
    --------
    Use the default Navarro geometry:

    >>> from visisipy import NavarroGeometry
    >>> geometry = NavarroGeometry()

    Create a Navarro geometry with a custom retina:

    >>> geometry = NavarroGeometry(
    ...     retina=StandardSurface(radius=-12.5, asphericity=0.5)
    ... )

    Create a default Navarro gometry and change only the lens back radius:

    >>> geometry = NavarroGeometry()
    >>> geometry.lens_back.radius = -5.8
    """

    def __init__(self, **kwargs):
        surfaces = {
            "cornea_front": StandardSurface(radius=7.72, asphericity=-0.26, thickness=0.55),
            "cornea_back": StandardSurface(radius=6.50, asphericity=0, thickness=3.05),
            "pupil": Stop(semi_diameter=1.348),
            "lens_front": StandardSurface(radius=10.2, asphericity=-3.1316, thickness=4.0),
            "lens_back": StandardSurface(radius=-6.0, asphericity=-1, thickness=16.3203),
            "retina": StandardSurface(radius=-12.0, asphericity=0),
        }
        surfaces.update(**kwargs)
        super().__init__(**surfaces)


def _update_attribute_if_specified(obj: Surface, attribute: str, value: Any):
    if value is not None:
        setattr(obj, attribute, value)


def _calculate_vitreous_thickness(geometry: EyeGeometry, parameters: GeometryParameters) -> float:
    """Calculate the thickness of the vitreous body for a partially initialized eye geometry."""

    # Axial length may be undefined, so parameters.get will not work here.
    _axial_length = parameters["axial_length"] if "axial_length" in parameters else geometry.axial_length
    _cornea_thickness = parameters.get("cornea_thickness", geometry.cornea_thickness)
    _anterior_chamber_depth = parameters.get("anterior_chamber_depth", geometry.anterior_chamber_depth)
    _pupil_lens_distance = parameters.get("pupil_lens_distance", geometry.pupil_lens_distance)
    _lens_thickness = parameters.get("lens_thickness", geometry.lens_thickness)

    if None in {
        _axial_length,
        _cornea_thickness,
        _anterior_chamber_depth,
        _pupil_lens_distance,
        _lens_thickness,
    }:
        raise ValueError("Cannot calculate vitreous thickness from the supplied parameters.")

    return _axial_length - (_cornea_thickness + _anterior_chamber_depth + _pupil_lens_distance + _lens_thickness)


GeometryType = TypeVar("GeometryType", bound=EyeGeometry)


class GeometryParameters(TypedDict, total=False):
    """Parameters for the geometry of the eye."""

    axial_length: float
    cornea_thickness: float
    cornea_front_radius: float
    cornea_front_asphericity: float
    cornea_back_radius: float
    cornea_back_asphericity: float
    anterior_chamber_depth: float
    pupil_radius: float
    pupil_lens_distance: float
    lens_thickness: float
    lens_back_radius: float
    lens_back_asphericity: float
    lens_front_radius: float
    lens_front_asphericity: float
    retina_radius: float
    retina_asphericity: float
    retina_ellipsoid_z_radius: float
    retina_ellipsoid_y_radius: float


def create_geometry(
    base: type[GeometryType] = NavarroGeometry,
    *,
    estimate_cornea_back: bool = False,
    **parameters: Unpack[GeometryParameters],
) -> GeometryType:
    """Create a geometry instance from clinically used parameters.

    All parameters are optional, and if not provided, the default values will be used.
    Sizes are specified in mm. If `estimate_cornea_back` is True, the back cornea radius will be estimated from the
    front cornea radius as `cornea_back_radius = 0.81 * cornea_front_radius`.
    The retina can be specified either by its radius and asphericity or by its axial and radial half axes. If both
    methods are specified, a ValueError will be raised.

    Parameters
    ----------
    base : type[GeometryType]
        The base geometry class to use. Must be a subclass of EyeGeometry.
    axial_length : float, optional
        Axial length of the eye, measured from cornea front to retina.
    cornea_thickness : float, optional
        Thickness of the cornea.
    cornea_front_radius : float, optional
        Radius of curvature of the frontal cornea surface.
    cornea_front_asphericity : float, optional
        Asphericity of the frontal cornea surface.
    cornea_back_radius : float, optional
        Radius of curvature of the back cornea surface.
    cornea_back_asphericity : float, optional
        Asphericity of the back cornea surface.
    anterior_chamber_depth : float, optional
        Depth of the anterior chamber.
    pupil_radius : float, optional
        Radius of the pupil.
    pupil_lens_distance : float, optional
        Distance between the pupil and the lens.
    lens_thickness : float, optional
        Thickness of the crystalline lens.
    lens_back_radius : float, optional
        Radius of curvature of the back lens surface.
    lens_back_asphericity : float, optional
        Asphericity of the back lens surface.
    lens_front_radius : float, optional
        Radius of curvature of the frontal lens surface.
    lens_front_asphericity : float, optional
        Asphericity of the frontal lens surface.
    retina_radius : float, optional
        Radius of curvature of the retina.
    retina_asphericity : float, optional
        Asphericity of the retina.
    retina_ellipsoid_z_radius : float, optional
        Radius (semi-axis length) of the retina ellipsoid in the z-direction.
    retina_ellipsoid_y_radius : float, optional
        Radius (semi-axis length) of the retina ellipsoid in the y-direction. For rotationally symmetric retinas,
        this is also the radius in the x-direction.
    estimate_cornea_back : bool, optional
        If True, the back cornea radius will be estimated from the front cornea radius. Default is `False`.

    Returns
    -------
    EyeGeometry
        An instance of the EyeGeometry class.

    Raises
    ------
    ValueError
        If the base geometry is not a class or if it is not a subclass of EyeGeometry.
        If the retina radius/asphericity and axial/radial half axes are both specified.
        If only one of the retina half axes is specified.
        If the sum of the cornea thickness, anterior chamber depth and lens thickness is greater than or equal to the
        axial length.
    """
    if not isinstance(base, type):
        raise TypeError("The base geometry must be a class. Did you put parentheses after the class name?")

    if not issubclass(base, EyeGeometry):
        raise TypeError("The base geometry must be a subclass of EyeGeometry.")

    if estimate_cornea_back and parameters.get("cornea_back_radius") is not None:
        warn("The cornea back radius was provided, but it will be ignored because estimate_cornea_back is True.")

    has_retina_radius_or_asphericity = ("retina_radius" in parameters) or ("retina_asphericity" in parameters)
    has_retina_ellipsoid_radii = ("retina_ellipsoid_z_radius" in parameters) or (
        "retina_ellipsoid_y_radius" in parameters
    )

    if has_retina_radius_or_asphericity and has_retina_ellipsoid_radii:
        raise ValueError("Cannot specify both retina radius/asphericity and ellipsoid radii.")

    if has_retina_ellipsoid_radii and not all(
        key in parameters for key in ("retina_ellipsoid_z_radius", "retina_ellipsoid_y_radius")
    ):
        raise ValueError("If the retina ellipsoid radii are specified, both the z and y radius must be provided.")

    geometry = base()

    _update_attribute_if_specified(geometry.cornea_front, "thickness", parameters.get("cornea_thickness"))
    _update_attribute_if_specified(geometry.cornea_front, "radius", parameters.get("cornea_front_radius"))
    _update_attribute_if_specified(geometry.cornea_front, "asphericity", parameters.get("cornea_front_asphericity"))

    if estimate_cornea_back:
        parameters["cornea_back_radius"] = 0.81 * geometry.cornea_front.radius

    _update_attribute_if_specified(geometry.cornea_back, "thickness", parameters.get("anterior_chamber_depth"))
    _update_attribute_if_specified(geometry.cornea_back, "radius", parameters.get("cornea_back_radius"))
    _update_attribute_if_specified(geometry.cornea_back, "asphericity", parameters.get("cornea_back_asphericity"))

    _update_attribute_if_specified(geometry.pupil, "semi_diameter", parameters.get("pupil_radius"))
    _update_attribute_if_specified(geometry.pupil, "thickness", parameters.get("pupil_lens_distance"))

    _update_attribute_if_specified(geometry.lens_front, "thickness", parameters.get("lens_thickness"))
    _update_attribute_if_specified(geometry.lens_front, "radius", parameters.get("lens_front_radius"))
    _update_attribute_if_specified(geometry.lens_front, "asphericity", parameters.get("lens_front_asphericity"))

    vitreous_thickness = _calculate_vitreous_thickness(geometry, parameters)

    if vitreous_thickness <= 0:
        raise ValueError(
            "The sum of the cornea thickness, anterior chamber depth, pupil-lens distance and lens thickness is "
            "greater than or equal to the axial length."
        )

    _update_attribute_if_specified(geometry.lens_back, "thickness", vitreous_thickness)
    _update_attribute_if_specified(geometry.lens_back, "radius", parameters.get("lens_back_radius"))
    _update_attribute_if_specified(geometry.lens_back, "asphericity", parameters.get("lens_back_asphericity"))

    if has_retina_radius_or_asphericity:
        _update_attribute_if_specified(geometry.retina, "radius", parameters.get("retina_radius"))
        _update_attribute_if_specified(geometry.retina, "asphericity", parameters.get("retina_asphericity"))
    elif has_retina_ellipsoid_radii:
        retina_radius = parameters["retina_ellipsoid_y_radius"] ** 2 / parameters["retina_ellipsoid_z_radius"]
        retina_asphericity = (
            parameters["retina_ellipsoid_y_radius"] ** 2 / parameters["retina_ellipsoid_z_radius"] ** 2 - 1
        )
        _update_attribute_if_specified(geometry.retina, "radius", retina_radius)
        _update_attribute_if_specified(geometry.retina, "asphericity", retina_asphericity)

    return geometry



"""Calculate the point spread function (PSF) of the eye."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, overload

from visisipy.analysis.base import _AUTOMATIC_BACKEND, analysis

if TYPE_CHECKING:
    from pandas import DataFrame

    from visisipy import EyeModel
    from visisipy.backend import BaseBackend
    from visisipy.types import FieldCoordinate, FieldType, SampleSize


@overload
def fft_psf(
    model: EyeModel | None = None,
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    sampling: SampleSize | str | int = 128,
    *,
    return_raw_result: Literal[False] = False,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> DataFrame: ...


@overload
def fft_psf(
    model: EyeModel | None = None,
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    sampling: SampleSize | str | int = 128,
    *,
    return_raw_result: Literal[True] = True,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> tuple[DataFrame, Any]: ...


@analysis
def fft_psf(
    model: EyeModel | None = None,  # noqa: ARG001
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    sampling: SampleSize | str | int = 128,
    *,
    return_raw_result: bool = False,  # noqa: ARG001
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> DataFrame | tuple[DataFrame, Any]:
    """Calculate the FFT Point Spread Function (PSF) at the retina surface.

    Parameters
    ----------
    model : EyeModel | None
        The eye model to be used in the ray trace. If `None`, the current eye model will be used.
    field_coordinate : FieldCoordinate | None
        The field coordinate at which the PSF is calculated. If `None`, the first field coordinate in the backend is
        used.
    wavelength : float | None
        The wavelength at which the PSF is calculated. If `None`, the first wavelength in the backend is used.
    field_type : FieldType
        The field type to be used in the analysis. Can be either "angle" or "object_height". Defaults to "angle".
        This parameter is only used when `field_coordinate` is specified.
    sampling : SampleSize | str | int
        The size of the ray grid used to sample the pupil. Can be an integer or a string in the format "NxN", where N
        is an integer. Only symmetric sample sizes are supported. Defaults to 128.
    return_raw_result : bool, optional
        Return the raw analysis result from the backend. Defaults to `False`.
    backend : type[BaseBackend]
        The backend to be used for the analysis. If not provided, the default backend is used.

    Returns
    -------
    DataFrame
        The PSF data as a DataFrame. The DataFrame contains the PSF values at the specified field coordinate and
        wavelength.
    """
    return backend.analysis.fft_psf(
        field_coordinate=field_coordinate,
        wavelength=wavelength,
        field_type=field_type,
        sampling=sampling,
    )


@overload
def huygens_psf(
    model: EyeModel | None = None,
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    pupil_sampling: SampleSize | str | int = 128,
    image_sampling: SampleSize | str | int = 128,
    *,
    return_raw_result: Literal[False] = False,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> DataFrame: ...


@overload
def huygens_psf(
    model: EyeModel | None = None,
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    pupil_sampling: SampleSize | str | int = 128,
    image_sampling: SampleSize | str | int = 128,
    *,
    return_raw_result: Literal[True] = True,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> tuple[DataFrame, Any]: ...


@analysis
def huygens_psf(
    model: EyeModel | None = None,  # noqa: ARG001
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    pupil_sampling: SampleSize | str | int = 128,
    image_sampling: SampleSize | str | int = 128,
    *,
    return_raw_result: bool = False,  # noqa: ARG001
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> DataFrame | tuple[DataFrame, Any]:
    """Calculate the Huygens Point Spread Function (PSF) at the retina surface.

    Parameters
    ----------
    model : EyeModel | None
        The eye model to be used in the analysis. If `None`, the current eye model will be used.
    field_coordinate : FieldCoordinate | None
        The field coordinate at which the PSF is calculated. If `None`, the first field coordinate in the backend is
        used.
    wavelength : float | None
        The wavelength at which the PSF is calculated. If `None`, the first wavelength in the backend is used.
    field_type : FieldType
        The field type to be used in the analysis. Can be either "angle" or "object_height". Defaults to "angle".
        This parameter is only used when `field_coordinate` is specified.
    pupil_sampling : SampleSize | str | int
        The size of the ray grid used to sample the pupil. Can be an integer or a string in the format "NxN", where N
        is an integer. Only symmetric sample sizes are supported. Defaults to 128.
    image_sampling : SampleSize | str | int
        The size of the PSF grid. Can be an integer or a string in the format "NxN", where N is an integer. Only
        symmetric sample sizes are supported. Defaults to 128.
    return_raw_result : bool, optional
        Return the raw analysis result from the backend. Defaults to `False`.
    backend : type[BaseBackend]
        The backend to be used for the analysis. If not provided, the default backend is used.

    Returns
    -------
    DataFrame
        The PSF data as a DataFrame. The DataFrame contains the PSF values at the specified field coordinate and
        wavelength.
    """
    return backend.analysis.huygens_psf(
        field_coordinate=field_coordinate,
        wavelength=wavelength,
        field_type=field_type,
        pupil_sampling=pupil_sampling,
        image_sampling=image_sampling,
    )


@overload
def strehl_ratio(
    model: EyeModel | None = None,
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    sampling: SampleSize | str | int = 128,
    psf_type: Literal["fft", "huygens"] = "huygens",
    *,
    return_raw_result: Literal[False] = False,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> float: ...


@overload
def strehl_ratio(
    model: EyeModel | None = None,
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    sampling: SampleSize | str | int = 128,
    psf_type: Literal["fft", "huygens"] = "huygens",
    *,
    return_raw_result: Literal[True] = True,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> tuple[float, Any]: ...


@analysis
def strehl_ratio(
    model: EyeModel | None = None,  # noqa: ARG001
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    sampling: SampleSize | str | int = 128,
    psf_type: Literal["fft", "huygens"] = "huygens",
    *,
    return_raw_result: bool = False,  # noqa: ARG001
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> float | tuple[float, Any]:
    """Calculate the Strehl ratio of the optical system.

    The Strehl ratio is calculated from the point spread function. Which PSF is used depends on the `psf_type` parameter.

    Parameters
    ----------
    model : EyeModel | None
        The eye model to be used in the ray trace. If `None`, the current eye model will be used.
    field_coordinate : FieldCoordinate | None
        The field coordinate at which the Strehl ratio is calculated. If `None`, the first field coordinate in the
        backend is used.
    wavelength : float | None
        The wavelength at which the Strehl ratio is calculated. If `None`, the first wavelength in the backend is used.
    field_type : FieldType
        The field type to be used in the analysis. Can be either "angle" or "object_height". Defaults to "angle".
        This parameter is only used when `field_coordinate` is specified.
    sampling : SampleSize | str | int
        The size of the ray grid used to sample the pupil. Can be an integer or a string in the format "NxN", where N
        is an integer. Only symmetric sample sizes are supported. Defaults to 128.
    psf_type : Literal["fft", "huygens"]
        The type of PSF to be used for the Strehl ratio calculation. Can be either "fft" or "huygens". Defaults to "huygens".
        Not all psf types are supported by all backends.
    return_raw_result : bool, optional
        Return the raw analysis result from the backend. Defaults to `False`.
    backend : type[BaseBackend]
        The backend to be used for the analysis. If not provided, the default backend is used.

    Returns
    -------
    float
        The Strehl ratio of the optical system at the specified field coordinate and wavelength.
    """
    if psf_type not in {"fft", "huygens"}:
        raise ValueError(f"Invalid psf_type: {psf_type}. Must be 'fft' or 'huygens'.")

    return backend.analysis.strehl_ratio(
        field_coordinate=field_coordinate,
        wavelength=wavelength,
        field_type=field_type,
        sampling=sampling,
        psf_type=psf_type,
    )



"""Calculate the spherical equivalent of refraction of the eye."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, overload

import numpy as np

from visisipy.analysis.base import _AUTOMATIC_BACKEND, analysis
from visisipy.refraction import FourierPowerVectorRefraction
from visisipy.types import SampleSize

if TYPE_CHECKING:
    from visisipy import EyeModel
    from visisipy.backend import BaseBackend, FieldCoordinate, FieldType
    from visisipy.wavefront import ZernikeCoefficients

__all__ = ("refraction", "zernike_data_to_refraction")


@overload
def refraction(
    model: EyeModel | None = None,
    field_coordinate: tuple[float, float] | None = None,
    sampling: SampleSize | str | int = 64,
    wavelength: float | None = None,
    pupil_diameter: float | None = None,
    field_type: Literal["angle", "object_height"] = "angle",
    *,
    use_higher_order_aberrations: bool = True,
    return_raw_result: Literal[False] = False,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> FourierPowerVectorRefraction: ...


@overload
def refraction(
    model: EyeModel | None = None,
    field_coordinate: tuple[float, float] | None = None,
    wavelength: float | None = None,
    sampling: SampleSize | str | int = 64,
    pupil_diameter: float | None = None,
    field_type: Literal["angle", "object_height"] = "angle",
    *,
    use_higher_order_aberrations: bool = True,
    return_raw_result: Literal[True] = True,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> tuple[FourierPowerVectorRefraction, Any]: ...


@analysis
def refraction(
    model: EyeModel | None = None,  # noqa: ARG001
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    sampling: SampleSize | str | int = 64,
    pupil_diameter: float | None = None,
    field_type: FieldType = "angle",
    *,
    use_higher_order_aberrations: bool = True,
    return_raw_result: bool = False,  # noqa: ARG001
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> tuple[FourierPowerVectorRefraction, Any]:
    """Calculates the ocular refraction.

    The ocular refraction is calculated from Zernike standard coefficients and represented in Fourier power vector form.

    Parameters
    ----------
    model : EyeModel, optional
        The eye model to use for the refraction calculation. When `None`, the currently built model is used.
        Defaults to `None`.
    use_higher_order_aberrations : bool, optional
        If `True`, higher-order aberrations are used in the calculation. Defaults to `True`.
    field_coordinate : tuple[float, float], optional
        The field coordinate for the Zernike calculation. When `None`, the default field configured in the backend is
        used. Defaults to `None`.
    wavelength : float, optional
        The wavelength for the Zernike calculation. When `None`, the default wavelength configured in the backend is
        used. Defaults to `None`.
    sampling : SampleSize | str | int, optional
        The sampling for the Zernike calculation. Defaults to 64.
    field_type : Literal["angle", "object_height"], optional
        The type of field to be used when setting the field coordinate. This parameter is only used when
        `field_coordinate` is specified. Defaults to "angle".
    pupil_diameter : float, optional
        The diameter of the pupil for the refraction calculation. Defaults to the pupil diameter configured in the
        model.
    return_raw_result : bool, optional
        Return the raw analysis result from the backend. Defaults to `False`.
    backend : type[BaseBackend]
        The backend to be used for the analysis. If not provided, the default backend is used.

    Returns
    -------
     FourierPowerVectorRefraction
          The ocular refraction in Fourier power vector form.
    Any
        The raw analysis result from the backend.

    See Also
    --------
    FourierPowerVectorRefraction : Ocular refraction in Fourier power vector form.
    """
    return backend.analysis.refraction(
        use_higher_order_aberrations=use_higher_order_aberrations,
        field_coordinate=field_coordinate,
        wavelength=wavelength,
        sampling=SampleSize(sampling),
        pupil_diameter=pupil_diameter,
        field_type=field_type,
    )


def zernike_data_to_refraction(
    zernike_coefficients: ZernikeCoefficients,
    exit_pupil_semi_diameter: float,
    wavelength: float,
    *,
    use_higher_order_aberrations: bool = True,
) -> FourierPowerVectorRefraction:
    """Convert Zernike coefficients to ocular refraction in Fourier power vector form.

    Parameters
    ----------
    zernike_coefficients
    exit_pupil_semi_diameter
    wavelength
    use_higher_order_aberrations

    Returns
    -------
    """
    z4 = zernike_coefficients[4] * wavelength * 4 * np.sqrt(3)
    z11 = zernike_coefficients[11] * wavelength * 12 * np.sqrt(5)
    z22 = zernike_coefficients[22] * wavelength * 24 * np.sqrt(7)
    z37 = zernike_coefficients[37] * wavelength * 40 * np.sqrt(9)

    z6 = zernike_coefficients[6] * wavelength * 2 * np.sqrt(6)
    z12 = zernike_coefficients[12] * wavelength * 6 * np.sqrt(10)
    z24 = zernike_coefficients[24] * wavelength * 12 * np.sqrt(14)
    z38 = zernike_coefficients[38] * wavelength * 60 * np.sqrt(2)

    z5 = zernike_coefficients[5] * wavelength * 2 * np.sqrt(6)
    z13 = zernike_coefficients[13] * wavelength * 6 * np.sqrt(10)
    z23 = zernike_coefficients[23] * wavelength * 12 * np.sqrt(14)
    z39 = zernike_coefficients[39] * wavelength * 60 * np.sqrt(2)

    if use_higher_order_aberrations:
        return FourierPowerVectorRefraction(
            M=(-z4 + z11 - z22 + z37) / (exit_pupil_semi_diameter**2),
            J0=(-z6 + z12 - z24 + z38) / (exit_pupil_semi_diameter**2),
            J45=(-z5 + z13 - z23 + z39) / (exit_pupil_semi_diameter**2),
        )

    return FourierPowerVectorRefraction(
        M=(-z4) / (exit_pupil_semi_diameter**2),
        J0=(-z6) / (exit_pupil_semi_diameter**2),
        J45=(-z5) / (exit_pupil_semi_diameter**2),
    )



"""Perform raytracing on an eye model."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, overload

from visisipy.analysis.base import _AUTOMATIC_BACKEND, analysis

if TYPE_CHECKING:
    from collections.abc import Iterable

    from pandas import DataFrame

    from visisipy import EyeModel
    from visisipy.backend import BaseBackend


@overload
def raytrace(
    model: EyeModel | None = None,
    coordinates: Iterable[tuple[float, float]] | None = None,
    wavelengths: Iterable[float] | None = None,
    field_type: Literal["angle", "object"] = "angle",
    pupil: tuple[float, float] = (0, 0),
    *,
    return_raw_result: Literal[False] = False,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> DataFrame: ...


@overload
def raytrace(
    model: EyeModel | None = None,
    coordinates: Iterable[tuple[float, float]] | None = None,
    wavelengths: Iterable[float] | None = None,
    field_type: Literal["angle", "object"] = "angle",
    pupil: tuple[float, float] = (0, 0),
    *,
    return_raw_result: Literal[True] = True,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> tuple[DataFrame, Any]: ...


@analysis
def raytrace(
    model: EyeModel | None = None,  # noqa: ARG001
    coordinates: Iterable[tuple[float, float]] | None = None,
    wavelengths: Iterable[float] | None = None,
    field_type: Literal["angle", "object"] = "angle",
    pupil: tuple[float, float] = (0, 0),
    *,
    return_raw_result: bool = False,  # noqa: ARG001
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> tuple[DataFrame, Any]:
    """Performs a ray trace analysis using the given parameters.
    The analysis returns a Dataframe with the following columns:

    - field: The field coordinates for the ray trace.
    - wavelength: The wavelength used in the ray trace.
    - surface: The surface number in the system.
    - comment: The comment for the surface.
    - x: The X-coordinate of the ray trace.
    - y: The Y-coordinate of the ray trace.
    - z: The Z-coordinate of the ray trace.

    Parameters
    ----------
    model : EyeModel | None
        The eye model to be used in the ray trace. If `None`, the current eye model will be used.
    coordinates : Iterable[tuple[float, float]], optional
        An iterable of tuples representing the coordinates for the ray trace.
        If `field_type` is "angle", the coordinates should be the angles along the (X, Y) axes in degrees.
        If `field_type` is "object_height", the coordinates should be the object heights along the
        (X, Y) axes in mm. Defaults to `None`, which uses the fields defined in the backend.
    wavelengths : Iterable[float], optional
        An iterable of wavelengths to be used in the ray trace. Defaults to `None`, which uses the wavelengths
        defined in the backend.
    field_type : Literal["angle", "object_height"], optional
        The type of field to be used in the ray trace. Can be either "angle" or "object_height". Defaults to "angle".
    pupil : tuple[float, float], optional
        A tuple representing the pupil coordinates for the ray trace. Defaults to (0, 0).
    return_raw_result : bool, optional
        Return the raw analysis result from the backend. Defaults to `False`.
    backend : type[BaseBackend]
        The backend to be used for the analysis. If not provided, the default backend is used.

    Returns
    -------
    DataFrame
        A pandas DataFrame containing the results of the ray trace analysis.
    Any
        The raw analysis result from the backend.
    """
    return backend.analysis.raytrace(
        coordinates=coordinates,
        wavelengths=wavelengths,
        field_type=field_type,
        pupil=pupil,
    )



from __future__ import annotations

import importlib
import platform
import sys
from importlib.machinery import ModuleSpec

__all__ = ("install_zospy_loader",)


class ZOSPyLoader:
    def __init__(self, name):
        self.name = name

    @staticmethod
    def load_module(fullname):
        if platform.system() != "Windows":
            message = (
                f"Could not import module '{fullname}'.\n"
                "Visisipy's OpticStudio backend depends on ZOSPy, which is not available on non-Windows systems. "
                "To use the OpticStudio backend, use Visisipy on a Windows system."
            )
            raise ImportError(message)

        # If on Windows, attempt to import the actual module
        return importlib.import_module(fullname)


class ZOSPyFinder:
    @staticmethod
    def find_spec(fullname, path, target=None) -> ModuleSpec | None:  # noqa: ARG004
        if fullname.split(".")[0] == "zospy":
            return ModuleSpec(name=fullname, loader=ZOSPyLoader(fullname))

        return None


def install_zospy_loader():
    """Install the ZOSPy loader to handle the import of the ZOSPy module."""
    if "zospy" not in sys.modules and not any(isinstance(m, ZOSPyFinder) for m in sys.meta_path):
        # Only add the finder if it's not already present
        # This prevents multiple finders from being added and causing import errors
        sys.meta_path.insert(0, ZOSPyFinder())



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from visisipy.opticstudio.backend import OpticStudioBackend


def set_wavelength(
    backend: type[OpticStudioBackend],
    wavelength: float | None = None,
) -> int:
    """Set the wavelength in the OpticStudio backend.

    Parameters
    ----------
    backend : OpticStudioBackend
        Reference to the OpticStudio backend.
    wavelength : float | None, optional
        The wavelength to set. If `None`, the first wavelength in the system will be used. Defaults to `None`.

    Returns
    -------
    int
        The wavelength number that was set.
    """
    wavelength_number = 1 if wavelength is None else backend.get_wavelength_number(wavelength)

    if wavelength_number is None:
        backend.set_wavelengths([wavelength])
        wavelength_number = 1

    return wavelength_number


def set_field(
    backend: type[OpticStudioBackend],
    field_coordinate: tuple[float, float] | None = None,
    field_type: str = "angle",
) -> int:
    """Set the field coordinate in the OpticStudio backend.

    Parameters
    ----------
    backend : OpticStudioBackend
        Reference to the OpticStudio backend.
    field_coordinate : tuple[float, float] | None, optional
        The field coordinate to set. If `None`, the first field in the system will be used. Defaults to `None`.
    field_type : str, optional
        The type of field to be used when setting the field coordinate. Defaults to "angle".

    Returns
    -------
    int
        The field number that was set.
    """
    if field_coordinate is not None:
        backend.set_fields([field_coordinate], field_type=field_type)

    return 1



"""OpticStudio backend for Visisipy."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, cast
from warnings import warn

import zospy as zp
from zospy.zpcore import OpticStudioSystem

from visisipy.backend import (
    BackendSettings,
    BaseBackend,
    _classproperty,
)
from visisipy.opticstudio.analysis import OpticStudioAnalysisRegistry
from visisipy.opticstudio.models import BaseOpticStudioEye, OpticStudioEye

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable
    from os import PathLike

    from zospy.api import _ZOSAPI
    from zospy.zpcore import ZOS

    from visisipy import EyeModel
    from visisipy.types import FieldCoordinate, NotRequired, Unpack


__all__ = ("OpticStudioBackend", "OpticStudioSettings")


RayAimingType = Literal["off", "paraxial", "real"]


class OpticStudioSettings(BackendSettings, total=False):
    """Backend settings that are specific to the OpticStudio backend."""

    mode: Literal["standalone", "extension"]
    """ZOSPy connection mode. Must be one of 'standalone' or 'extension'."""

    zosapi_nethelper: NotRequired[str]
    """Path to the ZOSAPI_NetHelper.dll file. If not provided, the path is automatically determined."""

    opticstudio_directory: NotRequired[str]
    """Path to the OpticStudio installation directory. If not provided, the path is automatically determined."""

    ray_aiming: RayAimingType
    """The ray aiming method to be used in the optical system. Must be one of 'off', 'paraxial', or 'real'."""


OPTICSTUDIO_DEFAULT_SETTINGS: OpticStudioSettings = {
    "field_type": "angle",
    "fields": [(0, 0)],
    "wavelengths": [0.543],
    "aperture_type": "float_by_stop_size",
    "mode": "standalone",
    "ray_aiming": "off",
}


class OpticStudioBackend(BaseBackend):
    """OpticStudio backend."""

    zos: ZOS | None = None
    oss: OpticStudioSystem | None = None
    model: BaseOpticStudioEye | None = None
    settings: OpticStudioSettings = OpticStudioSettings(**OPTICSTUDIO_DEFAULT_SETTINGS)
    _analysis: OpticStudioAnalysisRegistry | None = None

    @_classproperty
    def analysis(cls) -> OpticStudioAnalysisRegistry:  # noqa: N805
        """Provides access to the `OpticStudioAnalysisRegistry` instance.

        This property provides access to the `OpticStudioAnalysisRegistry` instance for performing various analyses on the optical
        system.

        Returns
        -------
        OpticStudioAnalysisRegistry
            The `OpticStudioAnalysisRegistry` instance.

        Raises
        ------
        RuntimeError
            If the OpticStudio backend has not been initialized.
        """
        if cls.oss is None:
            raise RuntimeError("The opticstudio backend has not been initialized.")
        if cls._analysis is None:
            cls._analysis = OpticStudioAnalysisRegistry(cls)

        return cls._analysis

    @classmethod
    def initialize(
        cls,
        **settings: Unpack[OpticStudioSettings],
    ) -> None:
        """Initialize the OpticStudio backend.

        This method connects to the OpticStudio backend and initializes a new optical system.

        Parameters
        ----------
        settings : OpticStudioSettings | None, optional
            The settings to be used for the OpticStudio backend. If None, the default settings are used.
        """
        if len(settings) > 0:
            cls.settings.update(settings)

        if cls.zos is None:
            cls.zos = zp.ZOS(
                zosapi_nethelper=cls.settings.get("zosapi_nethelper"),
                opticstudio_directory=cls.settings.get("opticstudio_directory"),
            )
        else:
            warn(
                "The OpticStudio backend has already been initialized. "
                "Reinitializing the backend is not necessary and may cause issues."
            )

        if cls.oss is None:
            cls.oss = cls.zos.connect(cls.get_setting("mode"))

        cls.new_model()

    @classmethod
    def update_settings(cls, **settings: Unpack[OpticStudioSettings]) -> None:
        """Apply the provided settings to the OpticStudio backend.

        This method applies the provided settings to the OpticStudio backend.
        """
        if len(settings) > 0:
            cls.settings.update(settings)

        if cls.oss is None:
            warn(
                "The OpticStudio backend settings can only be applied after initialization. "
                "Settings will be applied when the backend is initialized."
            )
        else:
            cls.set_aperture()
            cls.set_fields(cls.get_setting("fields"), field_type=cls.get_setting("field_type"))
            cls.set_ray_aiming(cls.oss, cls.get_setting("ray_aiming"))
            cls.set_wavelengths(cls.get_setting("wavelengths"))

    @classmethod
    def new_model(
        cls,
        *,
        save_old_model: bool = False,
    ) -> None:
        """Initialize a new optical system.

        This method initializes a new, empty optical system.
        """
        cls.get_oss().new(saveifneeded=save_old_model)

        cls.update_settings()

    @classmethod
    def build_model(
        cls,
        model: EyeModel,
        *,
        start_from_index: int = 0,
        replace_existing: bool = False,
        object_distance: float = float("inf"),
        **kwargs,
    ) -> OpticStudioEye:
        """Builds an optical system based on the provided eye model.

        This method creates an OpticStudioEye instance from the provided eye model and builds the optical system.
        If `replace_existing` is True, any existing model is updated instead of building a completely new system.

        Parameters
        ----------
        model : EyeModel
            The eye model to be used for building the optical system model.
        start_from_index : int, optional
            The index of the surface after which the eye model will be built. The cornea front surface will be located
            at `start_from_index + 1`.
        replace_existing : bool, optional
            Whether to replace any existing model before building the new one. Defaults to False.
        object_distance : float
            Distance between the cornea front and the surface preceding the eye model.
        **kwargs
            Additional keyword arguments to be passed to the OpticStudioEye build method.

        Returns
        -------
        OpticStudioEye
            The built optical system model.
        """
        if not replace_existing and cls.model is not None:
            cls.new_model()

        opticstudio_eye = OpticStudioEye(model)
        opticstudio_eye.build(
            cls.get_oss(),
            start_from_index=start_from_index,
            replace_existing=replace_existing,
            object_distance=object_distance,
            **kwargs,
        )

        cls.model = opticstudio_eye

        return opticstudio_eye

    @classmethod
    def clear_model(cls) -> None:
        """Clear the current optical system model.

        This method initializes a new optical system, discarding any existing model.
        """
        cls.get_oss().new(saveifneeded=False)
        cls.model = None

    @classmethod
    def save_model(cls, path: str | PathLike | None = None) -> None:
        """Save the current optical system model.

        This method saves the current optical system to the specified path. If no path is provided,
        it saves the model to the current working directory with the default name.

        Parameters
        ----------
        path : str | PathLike | None, optional
            The path where the model should be saved. If None, the model is saved in the current working directory.
        """
        if path is not None:
            cls.get_oss().save_as(path)
        else:
            cls.get_oss().save()

    @classmethod
    def disconnect(cls) -> None:
        """Disconnects the OpticStudio backend.

        This method closes the current optical system, sets the system and ZOS instances to None,
        and disconnects the ZOS instance.
        """
        if cls.oss is not None:
            cls.oss.close()
            cls.oss = None

        if cls.zos is not None:
            cls.zos.disconnect()
            cls.zos = None

    @classmethod
    def get_oss(cls) -> OpticStudioSystem:
        """Returns the current optical system.

        Returns
        -------
        OpticStudioSystem
            The current optical system.

        Raises
        ------
        RuntimeError
            If the OpticStudio system is not initialized.
        """
        if cls.oss is None:
            raise RuntimeError("No OpticStudio system initialized. Please initialize the backend first.")

        return cast(OpticStudioSystem, cls.oss)

    @classmethod
    def _set_aperture_value(cls) -> None:
        if cls.settings.get("aperture_value") is not None:
            cls.get_oss().SystemData.Aperture.ApertureValue = cls.get_setting("aperture_value")

    @classmethod
    def set_aperture(cls):
        if cls.get_setting("aperture_type") == "float_by_stop_size":
            cls.get_oss().SystemData.Aperture.ApertureType = zp.constants.SystemData.ZemaxApertureType.FloatByStopSize
        elif cls.get_setting("aperture_type") == "entrance_pupil_diameter":
            cls.get_oss().SystemData.Aperture.ApertureType = (
                zp.constants.SystemData.ZemaxApertureType.EntrancePupilDiameter
            )
            cls._set_aperture_value()
        elif cls.get_setting("aperture_type") == "image_f_number":
            cls.get_oss().SystemData.Aperture.ApertureType = zp.constants.SystemData.ZemaxApertureType.ImageSpaceFNum
            cls._set_aperture_value()
        elif cls.get_setting("aperture_type") == "object_numeric_aperture":
            cls.get_oss().SystemData.Aperture.ApertureType = zp.constants.SystemData.ZemaxApertureType.ObjectSpaceNA
            cls._set_aperture_value()
        else:
            raise ValueError(
                "aperture_type must be one of 'float_by_stop_size', 'entrance_pupil_diameter', "
                "'image_f_number', or 'object_numeric_aperture'."
            )

    @classmethod
    def get_fields(cls) -> list[FieldCoordinate]:
        """Get the fields in the optical system.

        Returns
        -------
        list[tuple[float, float]]
            The fields in the optical system as tuples of (x, y) coordinates.
        """
        fields = []

        for i in range(cls.get_oss().SystemData.Fields.NumberOfFields):
            field = cls.get_oss().SystemData.Fields.GetField(i + 1)
            fields.append((field.X, field.Y))

        return fields

    @staticmethod
    def _set_field_type(oss: OpticStudioSystem, field_type: Literal["angle", "object_height"]) -> None:
        if field_type == "angle":
            oss.SystemData.Fields.SetFieldType(zp.constants.SystemData.FieldType.Angle)
        elif field_type == "object_height":
            oss.SystemData.Fields.SetFieldType(zp.constants.SystemData.FieldType.ObjectHeight)
        else:
            raise ValueError("field_type must be either 'angle' or 'object_height'.")

    @staticmethod
    def _set_fields(
        oss: OpticStudioSystem,
        coordinates: Iterable[tuple[float, float]],
    ) -> None:
        oss.SystemData.Fields.DeleteAllFields()

        for i, c in enumerate(coordinates):
            if i == 0:
                field = oss.SystemData.Fields.GetField(1)
                field.X, field.Y, field.Weight = float(c[0]), float(c[1]), 1
            else:
                oss.SystemData.Fields.AddField(float(c[0]), float(c[1]), 1)

    @classmethod
    def set_fields(
        cls,
        coordinates: Iterable[tuple[float, float]],
        field_type: Literal["angle", "object_height"] = "angle",
    ) -> None:
        """Set the fields for the optical system.

        This method removes any existing fields and adds the new ones provided.

        Parameters
        ----------
        coordinates : Iterable[tuple[float, float]]
            An iterable of tuples representing the coordinates for the fields.
        field_type : Literal["angle", "object_height"], optional
            The type of field to be used in the optical system. Can be either "angle" or "object_height".
            Defaults to "angle".
        """
        cls._set_field_type(cls.get_oss(), field_type)
        cls._set_fields(cls.get_oss(), coordinates)

    @classmethod
    def get_wavelengths(cls) -> list[float]:
        """Get the wavelengths in the optical system.

        Returns
        -------
        list[float]
            The wavelengths in the optical system.
        """

        return [
            cls.get_oss().SystemData.Wavelengths.GetWavelength(i + 1).Wavelength
            for i in range(cls.get_oss().SystemData.Wavelengths.NumberOfWavelengths)
        ]

    @staticmethod
    def _remove_wavelenghts(oss: OpticStudioSystem) -> None:
        while oss.SystemData.Wavelengths.NumberOfWavelengths > 0:
            oss.SystemData.Wavelengths.RemoveWavelength(1)

    @classmethod
    def set_wavelengths(cls, wavelengths: Iterable[float]) -> None:
        """Set the wavelengths for the optical system.

        This method removes any existing wavelengths and adds the new ones provided.
        The weight for each wavelength is set to 1.0.

        Parameters
        ----------
        wavelengths : Iterable[float]
            An iterable of wavelengths to be set for the optical system.
        """
        cls._remove_wavelenghts(cls.get_oss())

        for w in wavelengths:
            cls.get_oss().SystemData.Wavelengths.AddWavelength(Wavelength=w, Weight=1.0)

    @classmethod
    def get_wavelength_number(cls, wavelength: float) -> int | None:
        """Returns the wavelength number for the given wavelength.

        If the wavelength is not found, `None` is returned.

        Parameters
        ----------
        wavelength : float
            The wavelength to find.

        Returns
        -------
        int | None
            The wavelength number, or `None` if the wavelength is not present.
        """
        for i in range(cls.get_oss().SystemData.Wavelengths.NumberOfWavelengths):
            if cls.get_oss().SystemData.Wavelengths.GetWavelength(i + 1).Wavelength == wavelength:
                return i + 1

        return None

    @staticmethod
    def set_ray_aiming(oss: OpticStudioSystem, ray_aiming: RayAimingType) -> None:
        if ray_aiming == "off":
            oss.SystemData.RayAiming.RayAiming = zp.constants.SystemData.RayAimingMethod.Off
        elif ray_aiming == "paraxial":
            oss.SystemData.RayAiming.RayAiming = zp.constants.SystemData.RayAimingMethod.Paraxial
        elif ray_aiming == "real":
            oss.SystemData.RayAiming.RayAiming = zp.constants.SystemData.RayAimingMethod.Real
        else:
            raise ValueError("ray_aiming must be either 'off', 'paraxial', or 'real'.")

    @classmethod
    def iter_fields(cls) -> Generator[tuple[int, _ZOSAPI.SystemData.IField], Any, None]:
        """Iterate over the fields in the optical system."""
        for i in range(cls.get_oss().SystemData.Fields.NumberOfFields):
            field = cls.get_oss().SystemData.Fields.GetField(i + 1)

            yield field.FieldNumber, field

    @classmethod
    def iter_wavelengths(cls) -> Generator[tuple[int, float], Any, None]:
        """Iterate over the wavelengths in the optical system."""
        for i in range(cls.get_oss().SystemData.Wavelengths.NumberOfWavelengths):
            yield i + 1, cls.get_oss().SystemData.Wavelengths.GetWavelength(i + 1).Wavelength



"""Refraction analysis for OpticStudio."""

from __future__ import annotations

from typing import TYPE_CHECKING

import zospy as zp

from visisipy.analysis.refraction import zernike_data_to_refraction
from visisipy.opticstudio.analysis.zernike_coefficients import (
    zernike_standard_coefficients,
)
from visisipy.types import SampleSize

if TYPE_CHECKING:
    from zospy.analyses.wavefront.zernike_standard_coefficients import ZernikeStandardCoefficientsResult

    from visisipy.opticstudio.backend import OpticStudioBackend
    from visisipy.refraction import FourierPowerVectorRefraction
    from visisipy.types import FieldCoordinate, FieldType


def refraction(
    backend: type[OpticStudioBackend],
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    sampling: SampleSize | str | int = 64,
    pupil_diameter: float | None = None,
    field_type: FieldType = "angle",
    *,
    use_higher_order_aberrations: bool = True,
) -> tuple[FourierPowerVectorRefraction, ZernikeStandardCoefficientsResult]:
    """Calculates the ocular refraction.

    The ocular refraction is calculated from Zernike standard coefficients and represented in Fourier power
    vector form.

    Parameters
    ----------
    backend : type[OpticStudioBackend]
        Reference to the OpticStudio backend.
    field_coordinate : tuple[float, float], optional
        The field coordinate for the Zernike calculation. When `None`, the first field in OpticStudio is used.
        Defaults to `None`.
    wavelength : float, optional
        The wavelength for the Zernike calculation. When `None`, the first wavelength in OpticStudio is used.
        Defaults to `None`.
    sampling : SampleSize | str | int, optional
        The sampling for the Zernike calculation. Defaults to 64.
    pupil_diameter : float, optional
        The diameter of the pupil for the refraction calculation. Defaults to the pupil diameter configured in the
        model.
    field_type : Literal["angle", "object_height"], optional
        The type of field to be used when setting the field coordinate. This parameter is only used when
        `field_coordinate` is specified. Defaults to "angle".
    use_higher_order_aberrations : bool, optional
        If `True`, higher-order aberrations are used in the calculation. Defaults to `True`.

    Returns
    -------
     FourierPowerVectorRefraction
          The ocular refraction in Fourier power vector form.
    """
    # Get the wavelength from OpticStudio if not specified
    wavelength = backend.get_wavelengths()[0] if wavelength is None else wavelength

    # Temporarily change the pupil diameter
    old_pupil_semi_diameter = None
    if pupil_diameter is not None:
        old_pupil_semi_diameter = backend.model.pupil.semi_diameter
        backend.model.pupil.semi_diameter = pupil_diameter / 2

    pupil_data = zp.functions.lde.get_pupil(backend.get_oss())
    zernike_coefficients, raw_result = zernike_standard_coefficients(
        backend,
        field_coordinate=field_coordinate,
        wavelength=wavelength,
        field_type=field_type,
        sampling=SampleSize(sampling),
    )

    if old_pupil_semi_diameter is not None:
        backend.model.pupil.semi_diameter = old_pupil_semi_diameter

    return zernike_data_to_refraction(
        zernike_coefficients,
        pupil_data.ExitPupilDiameter / 2,
        wavelength,
        use_higher_order_aberrations=use_higher_order_aberrations,
    ), raw_result



from __future__ import annotations

import math
from contextlib import nullcontext as does_not_raise
from types import SimpleNamespace

import numpy as np
import pytest
from zospy.solvers import material_model as solve_material_model

from visisipy.models.geometry import (
    NoSurface,
    StandardSurface,
    Stop,
    Surface,
    ZernikeStandardPhaseSurface,
    ZernikeStandardSagSurface,
)
from visisipy.models.materials import MaterialModel
from visisipy.opticstudio.surfaces import (
    BaseOpticStudioZernikeSurface,
    OpticStudioNoSurface,
    OpticStudioSurface,
    OpticStudioSurfaceDataProperty,
    OpticStudioSurfaceProperty,
    OpticStudioZernikeStandardPhaseSurface,
    OpticStudioZernikeStandardSagSurface,
    make_surface,
)
from visisipy.wavefront import ZernikeCoefficients

pytestmark = [pytest.mark.needs_opticstudio]


@pytest.fixture
def surface(oss):
    surface = oss.LDE.InsertNewSurfaceAt(1)
    surface.Comment = "Test comment"

    return surface


@pytest.mark.parametrize("surface_class", [OpticStudioSurface, OpticStudioZernikeStandardSagSurface])
def test_surface_type_exists(zos, surface_class: OpticStudioSurface):
    assert hasattr(zos.ZOSAPI.Editors.LDE.SurfaceType, surface_class._TYPE)


class TestOpticStudioSurfaceProperty:
    MockSurface = SimpleNamespace(
        Comment="Test comment",
    )

    class MockOpticStudioSurface:
        surface = None

        def __init__(self, surface):
            self.surface = surface

        comment = OpticStudioSurfaceProperty("Comment")

    def test_init(self):
        prop = OpticStudioSurfaceProperty("Comment")
        assert prop.name == "Comment"

    def test_get(self):
        mock_surface = self.MockOpticStudioSurface(self.MockSurface)

        assert mock_surface.comment == "Test comment"

    def test_get_none(self):
        mock_surface = self.MockOpticStudioSurface(None)

        assert mock_surface.comment is None

    def test_set(self):
        surface = self.MockSurface
        mock_surface = self.MockOpticStudioSurface(surface)

        mock_surface.comment = "New comment"

        assert mock_surface.comment == "New comment"
        assert surface.Comment == "New comment"

    def test_set_none(self):
        mock_surface = self.MockOpticStudioSurface(None)

        with pytest.raises(AttributeError, match="Cannot set attribute .+ of non-built surface"):
            mock_surface.comment = "New comment"


class TestOpticStudioSurfaceDataProperty:
    MockSurface = SimpleNamespace(
        SurfaceData=SimpleNamespace(
            ExampleProperty="Example value",
        )
    )

    class MockOpticStudioSurface:
        surface = None

        def __init__(self, surface):
            self.surface = surface

        example_property = OpticStudioSurfaceDataProperty("ExampleProperty")

    def test_init(self):
        prop = OpticStudioSurfaceDataProperty("ExampleProperty")
        assert prop.name == "ExampleProperty"

    def test_get(self):
        mock_surface = self.MockOpticStudioSurface(self.MockSurface)

        assert mock_surface.example_property == "Example value"

    def test_get_none(self):
        mock_surface = self.MockOpticStudioSurface(None)

        assert mock_surface.example_property is None

    def test_set(self, monkeypatch):
        surface = self.MockSurface
        mock_surface = self.MockOpticStudioSurface(surface)

        monkeypatch.setattr(mock_surface, "example_property", "New value")

        assert mock_surface.example_property == "New value"
        assert surface.SurfaceData.ExampleProperty == "New value"

    def test_set_none(self):
        mock_surface = self.MockOpticStudioSurface(None)

        with pytest.raises(AttributeError, match="Cannot set attribute .+ of non-built surface"):
            mock_surface.example_property = "New value"


class TestOpticStudioSurface:
    @pytest.mark.parametrize("is_stop", [True, None])
    def test_build(self, oss, is_stop):
        surface = OpticStudioSurface(
            comment="Test comment",
            radius=1.0,
            thickness=2.0,
            semi_diameter=3.0,
            conic=4.0,
            material="Test material",
            is_stop=is_stop,
        )

        assert surface._is_built is False
        surface_index = surface.build(oss, position=2)

        assert surface_index == 2
        assert surface._is_built is True

        assert surface.surface.SurfaceNumber == 2

        assert surface.comment == "Test comment"
        assert surface.radius == 1.0
        assert surface.thickness == 2.0
        assert surface.semi_diameter == 3.0
        assert surface.conic == 4.0
        assert surface.material == "TEST MATERIAL"  # OpticStudio capitalizes material names
        assert surface.is_stop is bool(is_stop)

    def test_is_stop_false_warns(self, oss):
        with pytest.warns(UserWarning, match="is_stop is set to False"):
            OpticStudioSurface(comment="Test", is_stop=False).build(oss, position=1)

    @pytest.mark.parametrize(
        "material_model",
        [
            MaterialModel(refractive_index=1.5, abbe_number=0, partial_dispersion=0),
            MaterialModel(refractive_index=1.5, abbe_number=50, partial_dispersion=0.67),
        ],
    )
    def test_set_material_model(self, oss, material_model):
        surface = OpticStudioSurface(comment="Test")
        surface.build(oss, position=1)
        surface.material = material_model

        material_solvedata = surface.surface.MaterialCell.GetSolveData()._S_MaterialModel
        assert material_solvedata.IndexNd == material_model.refractive_index
        assert material_solvedata.AbbeVd == material_model.abbe_number
        assert material_solvedata.dPgF == material_model.partial_dispersion

    @pytest.mark.parametrize(
        "refractive_index,abbe_number,partial_dispersion",
        [(1.5, 0, 0), (1.5, 50, 0.67)],
    )
    def test_get_material_model(self, oss, refractive_index, abbe_number, partial_dispersion):
        surface = OpticStudioSurface(comment="Test")
        surface.build(oss, position=1)
        solve_material_model(
            surface.surface.MaterialCell,
            refractive_index=refractive_index,
            abbe_number=abbe_number,
            partial_dispersion=partial_dispersion,
        )

        assert surface.material == MaterialModel(
            refractive_index=refractive_index,
            abbe_number=abbe_number,
            partial_dispersion=partial_dispersion,
        )

    def test_get_material_before_build(self):
        surface = OpticStudioSurface(comment="Test", material="Test material")

        assert surface.material is None

    def test_set_material_str(self, oss):
        surface = OpticStudioSurface(comment="Test")
        surface.build(oss, position=1)
        surface.material = "Test material"

        assert surface.surface.Material == "TEST MATERIAL"

    def test_get_material_str(self, oss):
        surface = OpticStudioSurface(comment="Test")
        surface.build(oss, position=1)
        surface.surface.Material = "Test material"

        assert surface.material == "TEST MATERIAL"

    def test_set_material_incorrect_type_raises_typeerror(self, oss):
        surface = OpticStudioSurface(comment="Test")
        surface.build(oss, position=1)

        with pytest.raises(TypeError, match="'material' must be MaterialModel or str"):
            surface.material = 5

    def test_relink_surface(self, oss):
        surface = OpticStudioSurface(comment="Test")
        surface.build(oss, position=2)

        assert surface.surface.SurfaceNumber == 2

        OpticStudioSurface(comment="New test").build(oss, position=1)

        assert surface.relink_surface(oss)
        assert surface.surface.SurfaceNumber == 3

    def test_relink_surface_changed_comment(self, oss):
        surface = OpticStudioSurface(comment="Test")
        surface.build(oss, position=2)
        surface._comment = "New test"

        assert not surface.relink_surface(oss)

    def test_set_surface_type(self, oss):
        surface = OpticStudioSurface(comment="Test")
        surface._TYPE = "ABCD"
        surface.build(oss, position=1)

        assert surface.surface.TypeName == "ABCD"

    def test_build_returns_index(self, oss):
        surface = OpticStudioSurface(comment="Test")

        assert surface.build(oss, position=1) == 1


class TestBaseOpticStudioZernikeSurface:
    class MockOpticStudioZernikeSurface(BaseOpticStudioZernikeSurface):
        """Very basic implementation of BaseOpticStudioZernikeSurface for testing purposes.

        This class is necessary because BaseOpticStudioZernikeSurface cannot be instantiated directly.
        """

        _TYPE = "ZernikeStandardSag"

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

    def test_instantiation_raises_typeerror(self):
        with pytest.raises(TypeError, match="Only child classes of BaseOpticStudioZernikeSurface may be instantiated"):
            BaseOpticStudioZernikeSurface("Useless comment")

    def test_init_large_zernike_coefficient_raises_valueerror(self):
        with pytest.raises(ValueError, match="Zernike coefficients must be smaller than the maximum term 1"):
            self.MockOpticStudioZernikeSurface(
                comment="Test",
                number_of_terms=1,
                zernike_coefficients={2: math.pi},
            )

    @pytest.mark.parametrize("key", [0, -1, -math.pi])
    def test_init_negative_zernike_coefficient_raises_valueerror(self, key):
        with pytest.raises(ValueError, match="Zernike coefficients must be larger than 0"):
            self.MockOpticStudioZernikeSurface(
                comment="Test",
                number_of_terms=3,
                zernike_coefficients={key: 1.234},
            )

    @pytest.mark.parametrize(
        "n,value,maximum_term,expectation",
        [
            (0, 1.234, 1, pytest.raises(ValueError, match="Zernike coefficient must be larger than 0")),
            (1, 1.234, 2, does_not_raise()),
            (2, 1.234, 2, does_not_raise()),
            (
                2,
                1.234,
                1,
                pytest.raises(ValueError, match="Zernike coefficient must be smaller than the maximum term 1"),
            ),
        ],
    )
    def test_set_zernike_coefficient(self, oss, n: int, value: float, maximum_term: int, expectation):
        surface = self.MockOpticStudioZernikeSurface(
            comment="Test",
            number_of_terms=maximum_term,
        )

        surface.build(oss, position=1)

        with expectation:
            surface.set_zernike_coefficient(n, value)
            assert surface.surface.SurfaceData.GetNthZernikeCoefficient(n) == value

    @pytest.mark.parametrize(
        "n,coefficients,expectation",
        [
            (2, {1: 1.234, 2: 3.456}, does_not_raise()),
            (
                4,
                {1: 1.234, 2: 3.456},
                pytest.raises(ValueError, match="Zernike coefficient must be smaller than the maximum term 3"),
            ),
            (0, {1: 1.234, 2: 3.456}, pytest.raises(ValueError, match="Zernike coefficient must be larger than 0")),
        ],
    )
    def test_get_zernike_coefficient(self, oss, n, coefficients, expectation):
        surface = OpticStudioZernikeStandardSagSurface(
            comment="Test",
            number_of_terms=3,
            zernike_coefficients=coefficients,
        )

        surface.build(oss, position=1)

        with expectation:
            assert surface.get_zernike_coefficient(n) == coefficients[n]


class TestOpticStudioZernikeStandardSagSurface:
    @pytest.mark.parametrize(
        "maximum_term,coefficients,extrapolate,decenter_x,decenter_y",
        [
            (3, {1: 1.0, 2: 2.0, 3: 3.0}, 1, 0.0, 0.0),
            (3, {1: 1.0, 2: 2.0, 3: 3.0}, 0, -1.0, 1.0),
            (3, {1: 1.0, 2: 2.0, 3: 3.0}, 1, 1.0, -1.0),
            (231, {}, 0, 0.0, 0.0),
            (231, None, 1, 0.0, 0.0),
        ],
    )
    def test_build(self, oss, maximum_term, coefficients, extrapolate, decenter_x, decenter_y):
        surface = OpticStudioZernikeStandardSagSurface(
            comment="Test comment",
            number_of_terms=maximum_term,
            extrapolate=extrapolate,
            zernike_decenter_x=decenter_x,
            zernike_decenter_y=decenter_y,
            zernike_coefficients=coefficients,
        )

        assert surface._is_built is False
        surface_index = surface.build(oss, position=1)

        assert surface_index == 1
        assert surface._is_built is True
        assert str(surface.surface.Type) == "ZernikeStandardSag"

        assert surface.comment == "Test comment"
        assert surface.number_of_terms == maximum_term
        assert surface.extrapolate == extrapolate
        assert surface.zernike_decenter_x == decenter_x
        assert surface.zernike_decenter_y == decenter_y

        for n, value in surface._zernike_coefficients.items():
            assert surface.get_zernike_coefficient(n) == value

    @pytest.mark.parametrize("decenter_x", np.arange(-2, 3, dtype=float))
    def test_zernike_decenter_x(self, oss, decenter_x):
        surface = OpticStudioZernikeStandardSagSurface(
            comment="Test",
            zernike_decenter_x=decenter_x,
        )

        surface.build(oss, position=1)

        surface.zernike_decenter_x = decenter_x
        assert surface.zernike_decenter_x == decenter_x
        assert surface.surface.SurfaceData.ZernikeDecenter_X == decenter_x

    @pytest.mark.parametrize("decenter_y", np.arange(-2, 3, dtype=float))
    def test_zernike_decenter_y(self, oss, decenter_y):
        surface = OpticStudioZernikeStandardSagSurface(
            comment="Test",
            zernike_decenter_y=decenter_y,
        )

        surface.build(oss, position=1)

        surface.zernike_decenter_y = decenter_y
        assert surface.zernike_decenter_y == decenter_y
        assert surface.surface.SurfaceData.ZernikeDecenter_Y == decenter_y

    @pytest.mark.parametrize("extrapolate", [0, 1, True, False])
    def test_extrapolate(self, oss, extrapolate):
        surface = OpticStudioZernikeStandardSagSurface(
            comment="Test",
            extrapolate=extrapolate,
        )

        surface.build(oss, position=1)

        surface.extrapolate = extrapolate
        assert surface.extrapolate == extrapolate
        assert surface.surface.SurfaceData.Extrapolate == extrapolate


class TestOpticStudioZernikeStandardPhaseSurface:
    @pytest.mark.parametrize(
        "maximum_term,coefficients,extrapolate,diffract_order",
        [
            (3, {1: 1.0, 2: 2.0, 3: 3.0}, 1, 1),
            (3, {1: 1.0, 2: 2.0, 3: 3.0}, 0, 2.3),
            (3, {1: 1.0, 2: 2.0, 3: 3.0}, 1, 3.4),
            (231, {}, 0, 4.5),
            (231, None, 1, 5.6),
        ],
    )
    def test_build(self, oss, maximum_term, coefficients, extrapolate, diffract_order):
        surface = OpticStudioZernikeStandardPhaseSurface(
            comment="Test comment",
            number_of_terms=maximum_term,
            extrapolate=extrapolate,
            diffract_order=diffract_order,
            zernike_coefficients=coefficients,
        )

        assert surface._is_built is False
        surface_index = surface.build(oss, position=1)

        assert surface_index == 1
        assert surface._is_built is True
        assert str(surface.surface.Type) == "ZernikeStandardPhase"

        assert surface.comment == "Test comment"
        assert surface.number_of_terms == maximum_term
        assert surface.extrapolate == extrapolate
        assert surface.diffract_order == diffract_order

        for n, value in surface._zernike_coefficients.items():
            assert surface.get_zernike_coefficient(n) == value

    @pytest.mark.parametrize("diffract_order", np.arange(0, 4, dtype=float))
    def test_diffract_order(self, oss, diffract_order):
        surface = OpticStudioZernikeStandardPhaseSurface(
            comment="Test",
            diffract_order=diffract_order,
        )

        surface.build(oss, position=1)

        surface.diffract_order = diffract_order
        assert surface.diffract_order == diffract_order
        assert surface.surface.SurfaceData.DiffractOrder == diffract_order

    @pytest.mark.parametrize("extrapolate", [0, 1, True, False])
    def test_extrapolate(self, oss, extrapolate):
        surface = OpticStudioZernikeStandardPhaseSurface(
            comment="Test",
            extrapolate=extrapolate,
        )

        surface.build(oss, position=1)

        surface.extrapolate = extrapolate
        assert surface.extrapolate == extrapolate
        assert surface.surface.SurfaceData.Extrapolate == extrapolate


class TestNoSurface:
    def test_build(self, oss):
        surface = OpticStudioNoSurface()

        n_surfaces = oss.LDE.NumberOfSurfaces

        return_index = surface.build(oss, position=1)

        assert return_index == 0
        assert surface.surface is None
        assert n_surfaces == oss.LDE.NumberOfSurfaces


class TestMakeSurface:
    def test_make_surface(self):
        surface = Surface(thickness=1)

        opticstudio_surface = make_surface(surface, material="BK7")

        assert opticstudio_surface._thickness == 1
        assert opticstudio_surface._material == "BK7"

    @pytest.mark.parametrize(
        "radius,thickness,semi_diameter,asphericity,material",
        [
            (1, 2, 3, 4, "BK7"),
            (
                5,
                6,
                7,
                8,
                MaterialModel(refractive_index=1.5, abbe_number=50, partial_dispersion=0.67),
            ),
        ],
    )
    def test_make_standard_surface(self, radius, thickness, semi_diameter, asphericity, material):
        surface = StandardSurface(
            radius=radius,
            thickness=thickness,
            semi_diameter=semi_diameter,
            asphericity=asphericity,
        )

        opticstudio_surface = make_surface(surface, material)

        assert opticstudio_surface._radius == radius
        assert opticstudio_surface._thickness == thickness
        assert opticstudio_surface._semi_diameter == semi_diameter
        assert opticstudio_surface._conic == asphericity
        assert opticstudio_surface._material == material

    @pytest.mark.parametrize(
        "thickness,semi_diameter,material",
        [
            (1, 2, "BK7"),
            (
                3,
                4,
                MaterialModel(refractive_index=1.5, abbe_number=50, partial_dispersion=0.67),
            ),
        ],
    )
    def test_make_stop_surface(self, thickness, semi_diameter, material):
        surface = Stop(thickness=thickness, semi_diameter=semi_diameter)

        opticstudio_surface = make_surface(surface, material)

        assert opticstudio_surface._thickness == thickness
        assert opticstudio_surface._semi_diameter == semi_diameter
        assert opticstudio_surface._material == material
        assert opticstudio_surface._is_stop is True

    def test_make_zernike_standard_sag_surface(self):
        surface = ZernikeStandardSagSurface(
            radius=1.0,
            thickness=2.0,
            semi_diameter=0.5,
            asphericity=0.1,
            zernike_decenter_x=0.2,
            zernike_decenter_y=0.3,
            maximum_term=3,
            norm_radius=50,
            extrapolate=False,
            zernike_coefficients=ZernikeCoefficients({1: 1.0, 2: 2.0, 3: 3.0}),
        )

        opticstudio_surface = make_surface(surface, material="BK7")

        assert opticstudio_surface._radius == 1.0
        assert opticstudio_surface._thickness == 2.0
        assert opticstudio_surface._semi_diameter == 0.5
        assert opticstudio_surface._conic == 0.1
        assert opticstudio_surface._zernike_decenter_x == 0.2
        assert opticstudio_surface._zernike_decenter_y == 0.3
        assert opticstudio_surface._number_of_terms == 3
        assert opticstudio_surface._norm_radius == 50
        assert opticstudio_surface._extrapolate == 0
        assert opticstudio_surface._zernike_coefficients == {1: 1.0, 2: 2.0, 3: 3.0}
        assert opticstudio_surface._material == "BK7"

    def test_make_zernike_standard_phase_surface(self):
        surface = ZernikeStandardPhaseSurface(
            radius=1.0,
            thickness=2.0,
            semi_diameter=0.5,
            asphericity=0.1,
            diffraction_order=2,
            maximum_term=3,
            norm_radius=50,
            extrapolate=True,
            zernike_coefficients=ZernikeCoefficients({1: 1.0, 2: 2.0, 3: 3.0}),
        )

        opticstudio_surface = make_surface(surface, material="BK7")

        assert opticstudio_surface._radius == 1.0
        assert opticstudio_surface._thickness == 2.0
        assert opticstudio_surface._semi_diameter == 0.5
        assert opticstudio_surface._conic == 0.1
        assert opticstudio_surface._diffract_order == 2
        assert opticstudio_surface._number_of_terms == 3
        assert opticstudio_surface._norm_radius == 50
        assert opticstudio_surface._extrapolate == 1
        assert opticstudio_surface._zernike_coefficients == {1: 1.0, 2: 2.0, 3: 3.0}
        assert opticstudio_surface._material == "BK7"

    def test_make_no_surface(self):
        surface = NoSurface()
        opticstudio_surface = make_surface(surface)

        assert isinstance(opticstudio_surface, OpticStudioNoSurface)



"""Build eye models in Optiland."""

from __future__ import annotations

from typing import TYPE_CHECKING

from visisipy.models import BaseEye
from visisipy.optiland.surfaces import OptilandSurface, make_surface

if TYPE_CHECKING:
    from optiland.optic import Optic

    from visisipy import EyeModel


__all__ = ("OptilandEye",)


class OptilandEye(BaseEye):
    """Eye model in Optiland."""

    def __init__(self, eye_model: EyeModel) -> None:
        """Create a new Optiland eye model.

        Parameters
        ----------
        eye_model : EyeModel
            Eye model specification frin which the Optiland eye model is created.
        """
        self._eye_model = eye_model

        self._cornea_front = make_surface(eye_model.geometry.cornea_front, eye_model.materials.cornea, "cornea front")
        self._cornea_back = make_surface(
            eye_model.geometry.cornea_back, eye_model.materials.aqueous, "cornea back / aqueous"
        )
        self._pupil = make_surface(eye_model.geometry.pupil, eye_model.materials.aqueous, "pupil")
        self._lens_front = make_surface(eye_model.geometry.lens_front, eye_model.materials.lens, "lens front")
        self._lens_back = make_surface(
            eye_model.geometry.lens_back,
            eye_model.materials.vitreous,
            "lens back / vitreous",
        )
        self._retina = make_surface(eye_model.geometry.retina, eye_model.materials.vitreous, "retina")

    @property
    def eye_model(self) -> EyeModel:
        """Eye model specification from which the Optiland eye model is created."""
        return self._eye_model

    @property
    def cornea_front(self) -> OptilandSurface:
        """Cornea front surface."""
        return self._cornea_front

    @property
    def cornea_back(self) -> OptilandSurface:
        """Cornea back surface."""
        return self._cornea_back

    @property
    def pupil(self) -> OptilandSurface:
        """Pupil surface."""
        return self._pupil

    @property
    def lens_front(self) -> OptilandSurface:
        """Lens front surface."""
        return self._lens_front

    @property
    def lens_back(self) -> OptilandSurface:
        """Lens back surface."""
        return self._lens_back

    @property
    def retina(self) -> OptilandSurface:
        """Retina surface."""
        return self._retina

    def build(
        self,
        optic: Optic,
        *,
        start_from_index: int = 0,
        replace_existing: bool = False,
        object_distance: float = float("inf"),
    ) -> None:
        """Create the eye in Optiland.

        Create the eye model in the provided optical system `optic`, starting from `start_from_index`.
        If `replace_existing` is set to `True`, existing surfaces will be overwritten.

        Parameters
        ----------
        optic : Optic
            Optiland Optic in which the eye model is created.
        start_from_index : int
            Index of the surface after which the eye model will be built.
        replace_existing : bool
            If `True`, replaces existing surfaces instead of inserting new ones. Defaults to `False`.
        object_distance : float, optional
            Distance from the object surface (or the surface before the eye model) to the eye model. Defaults to infinity.

        Raises
        ------
        AssertionError
            If the pupil is not located at the stop position.
            If the retina is not located at the last surface.
        """
        # Create an object surface if it does not exist
        if optic.surface_group.num_surfaces != 0 and object_distance != float("inf"):
            raise ValueError("Cannot set a custom object distance if the optical system is not empty.")

        if start_from_index == 0 and optic.surface_group.num_surfaces == 0:
            optic.surface_group.add_surface(
                index=start_from_index, replace_existing=replace_existing, thickness=object_distance
            )
        elif start_from_index > optic.surface_group.num_surfaces - 1:
            message = "'start_from_index' can be at most the index of the last surface in the system."
            raise ValueError(message)

        cornea_front_index = self.cornea_front.build(
            optic, position=start_from_index + 1, replace_existing=replace_existing
        )
        cornea_back_index = self.cornea_back.build(
            optic, position=cornea_front_index + 1, replace_existing=replace_existing
        )
        pupil_index = cornea_back_index + 1
        pupil_index = self.pupil.build(optic, position=pupil_index, replace_existing=replace_existing)
        lens_front_index = self.lens_front.build(optic, position=pupil_index + 1, replace_existing=replace_existing)
        lens_back_index = self.lens_back.build(optic, position=lens_front_index + 1, replace_existing=replace_existing)
        self.retina.build(optic, position=lens_back_index + 1, replace_existing=replace_existing)

        # Sanity checks
        if optic.surface_group.stop_index != pupil_index:
            message = "The pupil is not located at the stop position."
            raise ValueError(message)

        if optic.image_surface != self.retina.surface:
            message = "The retina is not located at the image position."
            raise ValueError(message)



"""Optical analyses for eye models.

This module provides access to various optical analyses for eye models.
The simulations themselves are implemented for each backend.
This module only provides a common interface for the analyses.

Analysis are functions with the following signature:

    analysis(model, [parameters], return_raw_result, backend)

The first parameter is the eye model to be used in the analysis. This parameter is optional;
if no model is provided, the model that is currently defined in the backend will be used.
Setting the return_raw_result parameter to True will return the raw analysis result from the backend.
The backend parameter is also optional; if no backend is provided, the current or default backend will be used.
"""

from __future__ import annotations

from visisipy.analysis.cardinal_points import cardinal_points
from visisipy.analysis.psf import fft_psf, huygens_psf, strehl_ratio
from visisipy.analysis.raytracing import raytrace
from visisipy.analysis.refraction import refraction
from visisipy.analysis.zernike_standard_coefficients import rms_hoa, zernike_standard_coefficients

__all__ = (
    "cardinal_points",
    "fft_psf",
    "huygens_psf",
    "raytrace",
    "refraction",
    "rms_hoa",
    "strehl_ratio",
    "zernike_standard_coefficients",
)



from __future__ import annotations

from contextlib import nullcontext as does_not_raise

import pytest

import visisipy
import visisipy.backend
from visisipy.wavefront import ZernikeCoefficients


class MockAnalysis:
    def cardinal_points(
        self,
        surface_1,
        surface_2,
    ):
        return None, None

    def fft_psf(self, field_coordinate, wavelength, field_type, sampling):
        return None, None

    def huygens_psf(
        self,
        field_coordinate,
        wavelength,
        field_type,
        pupil_sampling,
        image_sampling,
    ):
        return None, None

    def raytrace(
        self,
        coordinates,
        wavelengths,
        field_type,
        pupil,
    ):
        return None, None

    def refraction(
        self,
        field_coordinate,
        wavelength,
        sampling,
        pupil_diameter,
        field_type,
        *,
        use_higher_order_aberrations,
    ):
        return None, None

    def strehl_ratio(
        self,
        field_coordinate,
        wavelength,
        field_type,
        sampling,
        psf_type,
    ):
        return None, None

    def zernike_standard_coefficients(self, field_coordinate, wavelength, field_type, sampling, maximum_term):
        return ZernikeCoefficients({1: 0, 2: 0, 3: 0}), None


class MockBackend:
    analysis = MockAnalysis()
    model = object()


def test_cardinal_points_analysis(monkeypatch):
    monkeypatch.setattr(visisipy.backend, "_BACKEND", MockBackend)

    assert visisipy.analysis.cardinal_points() is None
    assert visisipy.analysis.cardinal_points(return_raw_result=True) == (None, None)


def test_fft_psf_analysis(monkeypatch):
    monkeypatch.setattr(visisipy.backend, "_BACKEND", MockBackend)

    assert visisipy.analysis.fft_psf() is None
    assert visisipy.analysis.fft_psf(return_raw_result=True) == (None, None)


def test_huygens_psf_analysis(monkeypatch):
    monkeypatch.setattr(visisipy.backend, "_BACKEND", MockBackend)

    assert visisipy.analysis.huygens_psf() is None
    assert visisipy.analysis.huygens_psf(return_raw_result=True) == (None, None)


def test_raytracing_analysis(monkeypatch):
    monkeypatch.setattr(visisipy.backend, "_BACKEND", MockBackend)

    assert visisipy.analysis.raytrace(coordinates=[(0, 0)]) is None
    assert visisipy.analysis.raytrace(coordinates=[(0, 0)], return_raw_result=True) == (
        None,
        None,
    )


def test_refraction_analysis(monkeypatch):
    monkeypatch.setattr(visisipy.backend, "_BACKEND", MockBackend)

    assert visisipy.analysis.refraction() is None
    assert visisipy.analysis.refraction(return_raw_result=True) == (None, None)


def test_rms_hoa_analysis(monkeypatch):
    monkeypatch.setattr(visisipy.backend, "_BACKEND", MockBackend)

    assert visisipy.analysis.rms_hoa() == 0
    assert visisipy.analysis.rms_hoa(return_raw_result=True) == (0, None)


def test_strehl_ratio_analysis(monkeypatch):
    monkeypatch.setattr(visisipy.backend, "_BACKEND", MockBackend)

    assert visisipy.analysis.strehl_ratio() is None
    assert visisipy.analysis.strehl_ratio(return_raw_result=True) == (None, None)


def test_zernike_standard_coefficients_analysis(monkeypatch):
    monkeypatch.setattr(visisipy.backend, "_BACKEND", MockBackend)

    assert visisipy.analysis.zernike_standard_coefficients() == ZernikeCoefficients({1: 0, 2: 0, 3: 0})
    assert visisipy.analysis.zernike_standard_coefficients(return_raw_result=True) == (
        ZernikeCoefficients({1: 0, 2: 0, 3: 0}),
        None,
    )


class TestRMSHOAAnalysis:
    @pytest.fixture(params=["opticstudio", "optiland"])
    def backend(self, request):
        fixture_name = request.param + "_backend"

        return request.getfixturevalue(fixture_name)

    @pytest.mark.parametrize(
        "min_order,max_order,maximum_term,expectation",
        [
            (3, 8, None, does_not_raise()),
            (
                -1,
                8,
                None,
                pytest.raises(ValueError, match="min_order and max_order must be greater than or equal to 0"),
            ),
            (
                0,
                -1,
                None,
                pytest.raises(ValueError, match="min_order and max_order must be greater than or equal to 0"),
            ),
            (0, 0, None, pytest.raises(ValueError, match="max_order must be greater than min_order")),
            (
                3,
                9,
                45,
                pytest.raises(
                    ValueError, match="maximum_term must be greater than or equal to the largest term of max_order"
                ),
            ),
        ],
    )
    def test_rms_hoa_analysis(self, min_order, max_order, maximum_term, expectation, backend, eye_model):
        backend.build_model(eye_model)

        with expectation:
            visisipy.analysis.rms_hoa(
                min_order=min_order,
                max_order=max_order,
                maximum_term=maximum_term,
                backend=backend,
            )


class TestStrehlRatioAnalysis:
    @pytest.mark.parametrize(
        "psf_type,expectation",
        [
            ("fft", does_not_raise()),
            ("huygens", does_not_raise()),
            (
                "invalid",
                pytest.raises(ValueError, match="Invalid psf_type: invalid. Must be 'fft' or 'huygens'"),
            ),
        ],
    )
    def test_psf_type(self, monkeypatch, psf_type, expectation):
        monkeypatch.setattr(visisipy.backend, "_BACKEND", MockBackend)

        with expectation:
            visisipy.analysis.strehl_ratio(psf_type=psf_type)
